"""Synchronous versions of common functions for optimized use cases."""
import itertools
from typing import Callable, Tuple

from gamla import construct, operator


def packstack(*functions):
    def packstack(values):
        return tuple(f(x) for f, x in zip(functions, values))

    return packstack


def keyfilter(predicate):
    def keyfilter(d):
        new_d = {}
        for k in d:
            if predicate(k):
                new_d[k] = d[k]
        return new_d

    return keyfilter


def valfilter(predicate):
    """Create a function that filters a dict using a predicate over values.
    >>> f = valefilter(lambda k: k > 2)
    >>> f({"a": 1, "b": 2, "c": 3, "d": 4})
    {
    "c": 3,
    "d": 4
    }
    """

    def valfilter(d):
        new_d = {}
        for (k, v) in d.items():
            if predicate(v):
                new_d[k] = d[k]
        return new_d

    return valfilter


def mapcat(f):
    def mapcat(it):
        for i in it:
            yield from f(i)

    return mapcat


def juxtcat(*functions):
    def juxtcat(x):
        for f in functions:
            for result in f(x):
                yield result

    return juxtcat


def map(f):
    def map_curried(it):
        for x in it:
            yield f(x)

    return map_curried


def filter(f):
    def filter(it):
        for x in it:
            if f(x):
                yield x

    return filter


def remove(f):
    def remove(it):
        for x in it:
            if not f(x):
                yield x

    return remove


def juxt(*functions):
    def juxt(*args, **kwargs):
        return tuple(f(*args, **kwargs) for f in functions)

    return juxt


def valmap(mapper):
    def valmap(d):
        new_d = {}
        for k in d:
            new_d[k] = mapper(d[k])
        return new_d

    return valmap


def keymap(mapper):
    def keymap(d):
        new_d = {}
        for k in d:
            new_d[mapper(k)] = d[k]
        return new_d

    return keymap


def groupby(grouper):
    def groupby(it):
        d = {}
        for x in it:
            key = grouper(x)
            if key not in d:
                d[key] = []
            d[key].append(x)
        return d

    return groupby


def groupby_many(grouper):
    def groupby_many(it):
        d = {}
        for x in it:
            for key in grouper(x):
                if key not in d:
                    d[key] = []
                d[key].append(x)
        return d

    return groupby_many


def ternary(condition, f_true, f_false):
    def ternary(*args, **kwargs):
        if condition(*args, **kwargs):
            return f_true(*args, **kwargs)
        return f_false(*args, **kwargs)

    return ternary


def check(condition, exception):
    """Apply function `condition` to value, raise `exception` if return value is `False`-ish or return the value as-is.

    >>> f = check(gamla.greater_than(10), ValueError)
    >>> f(5)
    `ValueError`
    >>> f(15)
    15
    """

    def check(x):
        if condition(x):
            return x
        raise exception

    return check


# TODO(uri): This might be used to optimize functions in gamla instead of its generic counterpart.
def compose(*functions):
    """Compose sync functions to operate in series.

    Returns a function that applies other functions in sequence.

    Functions are applied from right to left so that
    ``compose(f, g, h)(x, y)`` is the same as ``f(g(h(x, y)))``.

    >>> inc = lambda i: i + 1
    >>> compose(str, inc)(3)
    '4'

    """

    def compose(*args, **kwargs):
        for f in reversed(functions):
            args = [f(*args, **kwargs)]
            kwargs = {}
        return args[0]

    return compose


# TODO(uri): This might be used to optimize functions in gamla instead of its generic counterpart.
def compose_left(*functions):
    def compose_left(*args, **kwargs):
        for f in functions:
            x = f(*args, **kwargs)
            args = (x,)
            kwargs = {}
        return x

    return compose_left


# TODO(uri): This might be used to optimize functions in gamla instead of its generic counterpart.
def pipe(x, *functions):
    return compose_left(*functions)(x)


# TODO(uri): This might be used to optimize functions in gamla instead of its generic counterpart.
def anyjuxt(*functions):
    def anyjuxt(x):
        for f in functions:
            if f(x):
                return True
        return False

    return anyjuxt


def alljuxt(*functions):
    def alljuxt(x):
        for f in functions:
            if not f(x):
                return False
        return True

    return alljuxt


def complement(f):
    def complement(*args, **kwargs):
        return not f(*args, **kwargs)

    return complement


def allmap(f):
    def allmap(xs):
        for x in xs:
            if not f(x):
                return False
        return True

    return allmap


def anymap(f):
    def anymap(xs):
        for x in xs:
            if f(x):
                return True
        return False

    return anymap


# TODO(uri): This might be used to optimize functions in gamla instead of its generic counterpart.
def star(f):
    def starred(args):
        return f(*args)

    return starred


def double_star(f):
    def double_star(kwargs):
        return f(**kwargs)

    return double_star


def pair_left(f):
    def pair_left(x):
        return f(x), x

    return pair_left


def pair_right(f):
    """Returns a function that given a value x, returns a tuple of the form: (x, f(x)).

    >>> add_one = pair_right(lambda x: x + 1)
    >>> add_one(3)
    (3, 4)
    """

    def pair_right(x):
        return x, f(x)

    return pair_right


def reduce(f, initial):
    def reduce(it):
        state = initial
        for element in it:
            state = f(state, element)
        return state

    return reduce


def merge_with_reducer(reducer):
    def merge_with_reducer(*dictionaries):
        new_d = {}
        for d in dictionaries:
            for k, v in d.items():
                if k in new_d:
                    new_d[k] = reducer(new_d[k], v)
                else:
                    new_d[k] = v
        return new_d

    return merge_with_reducer


merge = star(merge_with_reducer(lambda _, x: x))


def after(f):
    def after(g):
        return compose_left(g, f)

    return after


def before(f):
    def before(g):
        return compose(g, f)

    return before


juxtduct = compose_left(juxt, after(star(itertools.product)))
mapdict = compose_left(map, after(dict))
mapduct = compose_left(map, after(star(itertools.product)))
maptuple = compose_left(map, after(tuple))


def binary_curry(f):
    def binary_curry(x):
        def binary_curry(y):
            return f(x, y)

        return binary_curry

    return binary_curry


def thunk(f, *args, **kwargs):
    def thunk(*inner_args, **inner_kwargs):
        return f(*args, **kwargs)(*inner_args, **inner_kwargs)

    return thunk


def when(f, g):
    def when(x):
        if f(x):
            return g(x)
        return x

    return when


class NoConditionMatched(Exception):  # noqa
    pass


def case(predicates_and_mappers: Tuple[Tuple[Callable, Callable], ...]):
    def case(*args, **kwargs):
        for predicate, transformation in predicates_and_mappers:
            if predicate(*args, **kwargs):
                return transformation(*args, **kwargs)
        raise NoConditionMatched({"input args": args, "input kwargs": kwargs})

    return case


case_dict = compose_left(dict.items, tuple, case)
stack = compose_left(
    enumerate,
    map(star(lambda i, f: compose(f, lambda x: x[i]))),
    star(juxt),
)

_is_terminal = anyjuxt(*map(operator.is_instance)([str, int, float]))


def map_dict(nonterminal_mapper: Callable, terminal_mapper: Callable) -> Callable:
    recurse = thunk(
        map_dict,
        nonterminal_mapper,
        terminal_mapper,
    )
    return case_dict(
        {
            _is_terminal: terminal_mapper,
            operator.is_instance(dict): compose_left(
                valmap(recurse),
                nonterminal_mapper,
            ),
            operator.is_iterable: compose_left(
                map(recurse),
                tuple,
                nonterminal_mapper,
            ),
            # Other types are considered terminals to support things like `apply_spec`.
            construct.just(True): terminal_mapper,
        },
    )


def bifurcate(*funcs):
    """Serially run each function on tee'd copies of a sequence.
    If the sequence is a generator, it is duplicated so it will not be exhausted (which may incur a substantial memory signature in some cases).
    >>> f = bifurcate(sum, gamla.count)
    >>> seq = map(gamla.identity, [1, 2, 3, 4, 5])
    >>> f(seq)
    (15, 5)
    """
    return compose_left(iter, lambda it: itertools.tee(it, len(funcs)), stack(funcs))



from typing import Any, Dict, Hashable


def wrap_dict(key: Hashable):
    """Wrap a key and a value in a dict (in a curried fashion).

    >>> wrap_dict("one") (1)
    {'one': 1}
    """

    def wrap_dict(value):
        return {key: value}

    return wrap_dict


def wrap_tuple(x: Any):
    """Wrap an element in a tuple.

    >>> wrap_tuple("hello")
    ('hello',)
    """
    return (x,)


def wrap_frozenset(x: Any):
    """Wraps x with frozenset.

    >>> wrap_frozenset(1)
    frozenset({1})
    """
    return frozenset([x])


def wrap_str(wrapping_string: str):
    """Wrap a string in a wrapping string.

    >>> wrap_str("hello {}", "world")
    'hello world'
    """

    def wrap_str(x) -> str:
        return wrapping_string.format(x)

    return wrap_str


def wrap_multiple_str(wrapping_string: str):
    """Wrap multiple values in a wrapping string by passing a dict where the keys are the parameters in the wrapping string and the values are the desired values.

    >>> wrap_multiple_str("hello {first} {second}")({ "first": "happy", "second": "world" })
    'hello happy world'
    """

    def inner(x: Dict[str, str]) -> str:
        return wrapping_string.format(**x)

    return inner


def just(x):
    """Ignores the input upon execution and returns the given argument.
    >>> f = just(1)
    >>> f(2)
    1
    """

    def just(*args, **kwargs):
        del args, kwargs
        return x

    return just



from gamla import url_utils


def test_add_to_query_string1():
    assert (
        url_utils.add_to_query_string(
            {"a": 123},
            "https://www.domain.com/path?param1=param1#anchor",
        )
        == "https://www.domain.com/path?param1=param1&a=123#anchor"
    )


def test_add_to_query_string2():
    assert (
        url_utils.add_to_query_string(
            {"param1": 123},
            "https://www.domain.com/path?param1=param1#anchor",
        )
        == "https://www.domain.com/path?param1=123#anchor"
    )



import asyncio
import functools
import logging
import time
from typing import Callable, Dict, Text, Tuple, Union

import async_timeout
import httpx
import requests
import requests.adapters
import termcolor
from requests.packages.urllib3.util import retry as retry_lib

from gamla import currying, functional
from gamla.optimized import sync


def _color_for_duration(duration: float):
    if duration < 0.1:
        return "green"
    if duration < 1:
        return "yellow"
    return "red"


def _log_finish(req_id: Text, start: float):
    finish = time.time()
    elapsed = finish - start
    logging.info(
        f"{termcolor.colored(req_id, 'white', 'on_grey')} {termcolor.colored(str(round(elapsed, 2)), _color_for_duration(elapsed))} seconds",
    )


def _async_timeit(f):
    @functools.wraps(f)
    async def wrapper(*args, **kwargs):
        start = time.time()
        result = await f(*args, **kwargs)
        _log_finish(f.__name__, start)
        return result

    return wrapper


def timeit(f):
    """Wraps a function `f` with a timer.
    Logs the start time, and end time (and difference in seconds).

    >>> timed_get_async = timeit(get_async)
    """
    if asyncio.iscoroutinefunction(f):
        return _async_timeit(f)

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        start = time.time()
        result = f(*args, **kwargs)
        _log_finish(f.__name__, start)
        return result

    return wrapper


def requests_with_retry(retries: int = 3) -> requests.Session:
    """Creates a `requests` object. Request will be attempted `retries` times.
    Will retry on 500, 502 and 504 status codes.

    >>> response = requests_with_retry(3).get("http://someurl.com")
    """
    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(
        max_retries=retry_lib.Retry(
            total=retries,
            backoff_factor=0.1,
            status_forcelist=(500, 502, 504),
        ),
    )
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session


@currying.curry
def batch_calls(max_batch_size: int, f: Callable):
    """Batches single call into one request.
    Turns `f`, a function that gets a `tuple` of independent requests, into a function
    that gets a single request.
    Each request will be at most of size `max_batched_size`.
    Requests will be performed in time intervals of 0.1s.

    >>> batched_f = batch_calls(5, f)
    """
    queue: Dict = {}

    async def make_call():
        await asyncio.sleep(0.1)
        if not queue:
            return
        slice_from_queue = tuple(functional.take(max_batch_size)(queue))
        promises = tuple(map(queue.__getitem__, slice_from_queue))
        requests = tuple(slice_from_queue)
        for x in slice_from_queue:
            del queue[x]
        try:
            for promise, result in zip(promises, await f(requests)):
                # We check for possible mid-exception or timeouts.
                if promise.done() or promise.cancelled():
                    continue
                promise.set_result(result)
        except Exception as exception:
            for promise in promises:
                # We check for possible mid-exception or timeouts.
                if promise.done() or promise.cancelled():
                    continue
                promise.set_exception(exception)

    @functools.wraps(f)
    async def wrapped(hashable_input):
        if hashable_input in queue:
            return await queue[hashable_input]
        async_result = asyncio.Future()
        # Check again because of context switch
        # due to the creation of `asyncio.Future`.
        # TODO(uri): Make sure this is needed.
        if hashable_input in queue:
            return await queue[hashable_input]
        queue[hashable_input] = async_result
        asyncio.create_task(make_call())
        return await async_result

    return wrapped


def queue_identical_calls(f):
    """Queues identical calls to coroutine `f` and performs each call once.
    Note that pending grows infinitely large, this assumes we cache the results
    anyway after this decorator, so we at most double the memory consumption.

    >>> deduped_get_async = queue_identical_calls(get_async)
    """

    pending = {}

    @functools.wraps(f)
    async def wrapped(*args, **kwargs):
        key = functional.make_call_key(args, kwargs)
        if key not in pending:
            pending[key] = asyncio.Future()
            pending[key].set_result(await f(*args, **kwargs))
        return await asyncio.wait_for(pending[key], timeout=20)

    return wrapped


def make_throttler(limit):
    """Returns a function that can be used on any number of coroutines to make sure only `limit` amount of calls are done in parallel.

    >>> throttler = make_throttler(3)
    >>> throttler(get_async)
    >>> throttler(post_json_async)
    """
    semaphore = None

    def wrap_function(f):
        @functools.wraps(f)
        async def wrap(*args, **kwargs):
            nonlocal semaphore
            # This must be in the inner function so that we avoid creating an event loop before the user has, causing the code to run with two different event loops.
            if not semaphore:
                semaphore = asyncio.Semaphore(limit)
            async with semaphore:
                return await f(*args, **kwargs)

        return wrap

    return wrap_function


@currying.curry
def throttle(limit, f):
    """Wraps a coroutine f assuring only `limit` amount of calls are done in parallel.

    >>> throttled_get_async = throttle(3, get_async)
    """
    return make_throttler(limit)(f)


def timeout(seconds: float):
    """Wraps a coroutine with a timeout (seconds) after
    which it will raise `asyncio.TimeoutError`.

    >>> get_with_timeout = timeout(3.5)(get_async)
    """

    def wrapper(corofunc):
        @functools.wraps(corofunc)
        async def run(*args, **kwargs):
            async with async_timeout.timeout(seconds):
                return await corofunc(*args, **kwargs)

        return run

    return wrapper


@currying.curry
async def get_async_with_headers(headers: Dict[str, str], timeout: float, url: Text):
    """Performs an async GET request to url with the specified timeout (seconds) and headers.

    >>> response = await get_async_with_headers({some_header: some_value}, 30, "https://www.someurl.com")
    """
    async with httpx.AsyncClient() as client:
        return await client.get(url, timeout=timeout, headers=headers)


#: Performs an async GET request to url with the specified timeout (seconds) and headers.
#: Expects the payload to be a json serializable object.
#:
#: >>> response = await get_async(30, "https://www.someurl.com")
get_async = get_async_with_headers({})


@currying.curry
async def head_async_with_headers(headers: Dict[str, str], timeout: float, url: Text):
    """Performs an async HEAD request to url with the specified timeout (seconds) and headers.

    >>> response = await head_async_with_headers(30, "https://www.someurl.com", {})
    """
    async with httpx.AsyncClient() as client:
        return await client.head(url, timeout=timeout, headers=headers)


@currying.curry
async def post_json_with_extra_headers_and_params_async(
    params: Dict[Text, Text],
    extra_headers: Dict[Text, Text],
    timeout: float,
    url: Text,
    payload,
):
    """Performs an http POST request with json data and URL parameters. Additional headers may be specified.
    Expects the params to be a dictionary object and the payload to be a json serializable object.

    >>> response = await post_json_with_extra_headers_and_params_async({"Date": "07/08/2022"}, {"Authorization": "Bearer TOKEN" }, 30, "https://www.someurl.com/post_data", { "name": "Danny" })
    """
    async with httpx.AsyncClient() as client:
        return await client.post(
            url=url,
            params=params,
            json=payload,
            headers=sync.merge(
                [
                    {"content_type": "application/json"},
                    extra_headers,
                ],
            ),
            timeout=timeout,
        )


#: Performs an http POST request of json data with specified headers.
#: Expects the payload to be a json serializable object and the headers to be a dictionary.
#:
#: >>> response = post_json_with_extra_headers_async({"username": "First Last", 30, "https://www.someurl.com/post_data", { "name": "Danny" })
post_json_with_extra_headers_async = post_json_with_extra_headers_and_params_async({})

#: Performs an http POST request of json data.
#: Expects the payload to be a json serializable object.
#:
#: >>> response = post_json_async(30, "https://www.someurl.com/post_data", { "name": "Danny" })
post_json_async = post_json_with_extra_headers_and_params_async({}, {})


@currying.curry
def retry(
    exception: Union[Exception, Tuple[Exception, ...]],
    times: int,
    wait_seconds: float,
    f: Callable,
):
    """Wraps a coroutine to retry on given exceptions."""

    async def retry_inner(*args, **kwargs):
        try:
            return await f(*args, **kwargs)
        except exception:
            await asyncio.sleep(wait_seconds)
            if not times:
                raise exception
            return await retry(exception, times - 1, wait_seconds, f)(*args, **kwargs)

    return retry_inner



import collections
import itertools
from typing import Callable, Iterable, Sequence, Union

concat = itertools.chain.from_iterable


def identity(x):
    return x


def nth(n):
    """Returns the nth element in a sequence.

    >>> nth(1, 'ABC')
    ['B']
    """

    def nth(seq):
        if isinstance(seq, (tuple, list, Sequence)):
            return seq[n]
        return next(itertools.islice(seq, n, None))

    return nth


def count(seq: Iterable) -> int:
    """Counts the number of items in `seq`. Similar to `len` but works on lazy sequences."""
    if hasattr(seq, "__len__"):
        return len(seq)  # type: ignore
    return sum(1 for _ in seq)


def attrgetter(attr):
    """Access the object attribute by its name `attr`.

    >>> attrgetter("lower")("ASD")()
    'asd'
    """

    def attrgetter(obj):
        return getattr(obj, attr)

    return attrgetter


def equals(x):
    def equals(y):
        return x == y

    return equals


def not_equals(x):
    """A functional !=.

    >>> not_equals(2)(2)
    False

    >>> not_equals("David")("Michael")
    True
    """

    def not_equals(y):
        return x != y

    return not_equals


def contains(x):
    """Contains operator.

    >>> contains([1, 2, 3])(2)
    True

    >>> contains("David")("x")
    False
    """

    def contains(y):
        return y in x

    return contains


def add(x):
    """Addition operator.

    >>> add(1)(2)
    3

    >>> add(["c"])(["a", "b"])
    ['a', 'b', 'c']
    """

    def add(y):
        return y + x

    return add


def greater_than(x):
    """Greater than operator.

    >>> greater_than(1)(2)
    True

    >>> greater_than(1)(0)
    False
    """

    def greater_than(y):
        return y > x

    return greater_than


def greater_equals(x):
    """Greater than or equal operator.

    >>> greater_equals(1)(1)
    True

    >>> greater_equals(1)(0)
    False
    """

    def greater_equals(y):
        return y >= x

    return greater_equals


def less_than(x):
    """Less than operator.

    >>> less_than(1)(1)
    False
    """

    def less_than(y):
        return y < x

    return less_than


def less_equals(x):
    """Less than or equal operator.

    >>> less_equals(1)(1)
    True

    >>> less_equals(1)(3)
    False
    """

    def less_equals(y):
        return y <= x

    return less_equals


def multiply(x):
    """Multiply operator.

    >>> multiply(2)(1)
    2
    """

    def multiply(y):
        return y * x

    return multiply


def divide_by(x):
    def divide_by(y):
        return y / x

    return divide_by


def inside(val):
    """A functional `in` operator.

    >>> inside(1)([0, 1, 2])
    True

    >>> inside("a", "word")
    False
    """

    def inside(container):
        return val in container

    return inside


def len_equals(length: int):
    """Measures if the length of a sequence equals to a given length.

    >>> len_equals(3)([0, 1, 2])
    True
    """

    def len_equals(x: Iterable) -> bool:
        return count(x) == length

    return len_equals


def len_greater(length: int):
    """Measures if the length of a sequence is greater than a given length.

    >>> len_greater(2)([0, 1, 2])
    True
    """

    def len_greater(seq):
        return count(seq) > length

    return len_greater


def len_smaller(length: int) -> Callable:
    """Measures if the length of a sequence is smaller than a given length.

    >>> len_smaller(2)([0, 1, 2])
    False
    """

    def len_smaller(seq):
        return count(seq) < length

    return len_smaller


def between(low: int, high: int):
    def between(number: Union[int, float]):
        return low <= number < high

    return between


def empty(seq):
    try:
        next(iter(seq))
    except StopIteration:
        return True
    return False


def nonempty(seq):
    return not empty(seq)


def head(seq):
    """Returns the first element in a sequence.
    >>> first('ABC')
    'A'
    """
    return next(iter(seq))


def second(seq):
    """Returns the second element in a sequence.
    >>> second('ABC')
    'B'
    """
    seq = iter(seq)
    next(seq)
    return next(seq)


def tail(n):
    """Returns the last n elements of a sequence.
    >>> tail(2, [10, 20, 30, 40, 50])
    [40, 50]
    """

    def tail(seq):
        try:
            return seq[-n:]
        except (TypeError, KeyError):
            return tuple(collections.deque(seq, n))

    return tail


def last(seq):
    """Returns the last element in a sequence
    >>> last('ABC')
    'C'
    """
    return tail(1)(seq)[0]


def pack(*stuff):
    """Returns a list generated from the provided input.

    >>> pack(1, 2, 3)
    (1, 2, 3)
    """
    return stuff


def is_instance(the_type):
    """Returns if `the_value` is an instance of `the_type`.

    >>> is_instance(str)("hello")
    True

    >>> is_instance(int)("a")
    False
    """

    def is_instance(the_value):
        return isinstance(the_value, the_type)

    return is_instance


def is_iterable(x):
    """Determines whether the element is iterable.

    >>> isiterable([1, 2, 3])
    True
    >>> isiterable('abc')
    True
    >>> isiterable(5)
    False"""
    try:
        iter(x)
        return True
    except TypeError:
        return False



"""Asynchronous versions of common functions for optimized use cases."""
import asyncio
import inspect


async def to_awaitable(value):
    """Wraps a value in a coroutine.
    If value is a future, it will await it. Otherwise it will simply return the value.
    Useful when we have a mix of coroutines and regular functions.

    >>> run_sync(await to_awaitable(5))
    '5'
    >>> run_sync(await to_awaitable(some_coroutine_that_returns_its_input(5)))
    '5'
    """
    if inspect.isawaitable(value):
        return await value
    return value


def compose_left(*funcs):
    async def compose_left_async(*args, **kwargs):
        for f in funcs:
            args = [await to_awaitable(f(*args, **kwargs))]
            kwargs = {}
        return args[0]

    return compose_left_async


def compose(*funcs):
    async def compose(*args, **kwargs):
        for f in reversed(funcs):
            args = [await to_awaitable(f(*args, **kwargs))]
            kwargs = {}
        return args[0]

    return compose


def map(f):
    async def map(it):
        return await asyncio.gather(*(f(x) for x in it))

    return map


def star(f):
    async def star(x):
        return await f(*x)

    return star


def double_star(f):
    async def double_star(x):
        return await f(**x)

    return double_star


def thunk(f, *args, **kwargs):
    async def thunk(*inner_args, **inner_kwargs):
        return await f(*args, **kwargs)(*inner_args, **inner_kwargs)

    return thunk



from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    FrozenSet,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import pytest

from gamla import type_safety


def test_no_annotations():
    def f(x):
        pass

    def g(x):
        pass

    assert type_safety.composable(f, g, None)


def test_simple_types():
    def f(x: int) -> str:
        pass

    def g(x: str) -> str:
        pass

    assert type_safety.composable(g, f, None)
    assert not type_safety.composable(f, g, None)


def test_compose_on_key():
    def f(x: int, y: str) -> str:
        pass

    def g(x: str) -> str:
        pass

    assert type_safety.composable(f, g, "y")
    assert not type_safety.composable(f, g, "x")


_T = TypeVar("_T")
_T_number = TypeVar("_T_number", int, float)
_T_collection_or_dict_of_strings = TypeVar(
    "_T_collection_or_dict_of_strings",
    Dict[str, str],
    Collection[str],
)
_T_list_or_frozenset_of_strings = TypeVar(
    "_T_list_or_frozenset_of_strings",
    List[str],
    FrozenSet[str],
)


@pytest.mark.parametrize(
    "subtype,supertype",
    [
        [Callable, Callable],
        [Callable[[str], str], Callable],
        [Callable[..., str], Callable],
        [Callable[[str], str], Callable[[str], str]],
        [Callable[..., str], Callable[[str], str]],
        [Callable[[str], str], Callable[..., str]],
        [Dict[str, int], Dict[str, int]],
        [type(None), Optional[str]],
        [type(None), type(None)],
        [Optional[str], Optional[str]],
        [Optional[str], Union[None, str]],
        [str, Optional[str]],
        [FrozenSet[str], FrozenSet[str]],
        [str, Any],
        [Tuple[str, ...], Tuple[str, ...]],
        [Set[str], Collection[str]],
        [List, Sequence],
        [Union[int, str], Union[int, str]],
        [str, Union[int, str]],
        [Union[List, Set], Collection],
        [Dict, Dict[str, int]],
        [FrozenSet[int], Collection[_T]],
        [FrozenSet[int], Collection[_T_number]],
        [_T_list_or_frozenset_of_strings, _T_collection_or_dict_of_strings],
    ],
)
def test_is_subtype(subtype, supertype):
    assert type_safety.is_subtype(subtype, supertype)


@pytest.mark.parametrize(
    "subtype,supertype",
    [
        [Callable, Callable[[str], str]],
        [Callable, Callable[..., str]],
        [Callable[[str, str], str], Callable[[str], str]],
        [Callable[[int, str], str], Callable[[str, int], str]],
        [Dict[str, str], Dict[str, int]],
        [Optional[str], type(None)],
        [Optional[str], Optional[int]],
        [int, Optional[str]],
        [FrozenSet[int], FrozenSet[str]],
        [str, FrozenSet[str]],
        [Collection, FrozenSet],
        [Tuple[str, ...], Tuple[int, ...]],
        [Union[int, str], int],
        [Any, str],
        [List, Union[int, str]],
        [Union[int, str, List], Union[int, str]],
        [Dict, Tuple],
        [FrozenSet[str], Collection[_T_number]],
    ],
)
def test_is_not_subtype(subtype, supertype):
    assert not type_safety.is_subtype(subtype, supertype)



from gamla import dict_utils, functional_generic, operator


def test_itemgetter():
    assert dict_utils.itemgetter("a")({"a": 1}) == 1


def test_itemgetter_with_default():
    assert dict_utils.itemgetter_with_default(2, "b")({"a": 1}) == 2


def test_itemgetter_or_none():
    assert dict_utils.itemgetter_or_none("b")({"a": 1}) is None


def test_get_in():
    assert dict_utils.get_in(["a", "b", "c", 1])({"a": {"b": {"c": [0, 1, 2]}}}) == 1


def test_get_in_or_none():
    assert (
        dict_utils.get_in_or_none(["a", "b", "d", 1])({"a": {"b": {"c": [0, 1, 2]}}})
        is None
    )


def test_get_in_or_none_uncurried():
    assert (
        dict_utils.get_in_or_none_uncurried(
            ["a", "b", "c", 1],
            {"a": {"b": {"c": [0, 1, 2]}}},
        )
        == 1
    )


def test_dict_to_getter_with_default_value_exists():
    assert dict_utils.dict_to_getter_with_default(None, {1: 1})(1) == 1


def test_dict_to_getter_with_default_values_does_not_exist():
    assert dict_utils.dict_to_getter_with_default(None, {1: 1})(2) is None


def test_get_or_transform():
    assert dict_utils.get_or_transform(operator.add(1), {1: 1})(1) == 1
    assert dict_utils.get_or_transform(operator.add(1), {1: 1})(3) == 4


def test_get_or_identity():
    assert dict_utils.get_or_identity({1: 1})(1) == 1
    assert dict_utils.get_or_identity({1: 1})(2) == 2


def test_make_index():
    index = dict_utils.make_index(
        map(functional_generic.groupby, [operator.head, operator.second]),
    )(["uri", "dani"])
    assert index("d")("a") == frozenset(["dani"])
    assert index("u")("r") == frozenset(["uri"])
    assert index("h")("i") == frozenset()


def test_rename_key():
    assert dict_utils.rename_key("name", "first_name")(
        {"name": "Danny", "age": 20},
    ) == {"first_name": "Danny", "age": 20}


def test_add_key_value():
    assert dict_utils.add_key_value("1", "1")({"2": "2"}) == {"1": "1", "2": "2"}


def test_remove_key():
    assert dict_utils.remove_key("1")({"1": 1, "2": 2}) == {"2": 2}


def test_transform_item():
    assert dict_utils.transform_item("1", operator.add(1))({"1": 1, "2": 2}) == {
        "1": 2,
        "2": 2,
    }



from gamla import operator
from gamla.optimized import sync


def test_valfilter_sync():
    assert sync.valfilter(operator.identity)(
        {False: True, True: False},
    ) == {False: True}


def test_pair_right_sync():
    assert sync.pair_right(operator.multiply(2))(4) == (4, 8)



import asyncio
from typing import Any, AsyncGenerator, Awaitable, Callable, Iterable, Type

from gamla import currying
from gamla.optimized import async_functions


def run_sync(f):
    """Runs a coroutine in a synchronous context, blocking until result arrives.

    >>> async def afoo(x):
    ...     await asyncio.sleep(1)
    ...     return x
    ... run_sync(afoo(1))
    1 (after 1 second of waiting)
    """
    loop = asyncio.new_event_loop()
    return loop.run_until_complete(asyncio.ensure_future(f, loop=loop))


@currying.curry
async def amap_ascompleted(
    f: Callable[[Any], Awaitable[Any]],
    it: Iterable,
) -> AsyncGenerator[Any, None]:
    """Returns an AsyncGenerator of the results after applying async `f` to each element of Iterable `it`

    >>> async def amulti(x)
    ...    return x*2
    ... async def to_list(ag):
    ...     return [i async for i in ag]
    ... run_sync(to_list(amap_ascompleted(amulti, range(4))))
    [6, 0, 2, 4] (In a random order)
    """
    for future in asyncio.as_completed(map(f, it)):
        yield await future


@currying.curry
async def aexcepts(
    exception_type: Type[Exception],
    func: Callable,
    handler: Callable,
    x: Any,
):
    """An async functional try/except block: await and return `func` on `x`.
    If fails with `exception_type`, return the reult of running handler on the error.


    >>> async def araise(x):
    ...     raise ValueError
    ... run_sync(aexcepts(ValueError, araise, lambda e: e, 5))
    ValueError()

    >>> async def a_just(x):
    ...     return x
    ... run_sync(aexcepts(ValueError, a_just, lambda e: e, 5))
    5
    """
    try:
        return await func(x)
    except exception_type as error:
        return handler(error)


@currying.curry
async def mapa(f: Callable, it: AsyncGenerator) -> AsyncGenerator:
    """Returns an AsyncGenerator of the results after applying `f` to each async element of `it`

    >>> async def arange(count):
    ...     for i in range(count):
    ...         yield(i)
    ... async def to_list(ag):
    ...     return [i async for i in ag]
    ... run_sync(to_list(mapa(lambda x: x*2, arange(4))))
    [0, 2, 4, 6]
    """
    async for element in it:
        yield f(element)


async def aconcat(async_generator: AsyncGenerator) -> AsyncGenerator:
    """Concat iterables of an async_generator.

    >>> async def many_range(count):
    ...     for i in range(count):
    ...         yield range(i, i+1)
    ... async def to_list(ag):
    ...     return [i async for i in ag]
    ... run_sync(to_list(aconcat(many_range(4))))
    [0, 1, 2, 3]
    """
    async for g in async_generator:
        for x in g:
            yield x


def afirst(*funcs: Callable, exception_type: Type[Exception]):
    """Runs given `funcs` serially until getting a succefull result.
    Returns the result of the first function that runs on `x` without raising `exception_type`.
    If all given function raise e`xception_type`, `exception_type` will be raised.

    >>> async def araise(x):
    ...     raise ValueError
    ... run_sync(afirst(araise, lambda x: x*x, exception_type=ValueError)(3))
    9
    """

    async def afirst_inner(x: Any):
        for f in funcs:
            try:
                return await async_functions.to_awaitable(f(x))
            except exception_type:
                pass
        raise exception_type

    return afirst_inner



"""https://www.youtube.com/watch?v=6mTbuzafcII for some explanations about this is about."""

import functools
import itertools
from typing import Any, Callable, Iterable, TypeVar

from gamla import construct, dict_utils, functional_generic, operator
from gamla.optimized import sync

_S = TypeVar("_S")
_X = TypeVar("_X")
Reducer = Callable[[_S, _X], _S]
Transducer = Callable[[Reducer], Reducer]


def transduce(transformation: Transducer, step: Reducer, initial, collection):
    """Reduces over a `collection` using a `step` function to get elements and a `transformation` logic.

    See other transducer functions for usage examples."""
    return functools.reduce(transformation(step), collection, initial)


def _transform_by_key(injector):
    return lambda key, reducer: lambda step: lambda s, x: step(
        injector(s, key, reducer(s[key], x)),
        x,
    )


def _replace_index(the_tuple, index, value):
    return (*the_tuple[:index], value, *the_tuple[index + 1 :])


#: Combines transducers in a `dict` into a transducer that produces a `dict`.
#: >>> transducer.transduce(
#: ...     transducer.apply_spec({  # This will combine the inner stuff into one new transducer.
#: ...         "incremented": _increment(_append_to_tuple),  # This is a transducer.
#: ...         "sum": lambda s, x: x + s,  # This is another transducer.
#: ...     }),
#: ...     lambda s, _: s,
#: ...     {"incremented": (), "sum": 0},
#: ...     [1, 2, 3],
#: ... )
#: {"incremented": (2, 3, 4), "sum": 6}
apply_spec = functional_generic.compose_left(
    dict.items,
    sync.map(
        sync.star(_transform_by_key(lambda x, y, z: dict_utils.add_key_value(y, z)(x))),
    ),
    sync.star(functional_generic.compose),
)

#: Combines transducers in a `tuple` into a transducer that produces a `tuple`.
#: >>> transducer.transduce(
#: ...     transducer.juxt(  # This will combine the inner stuff into one new transducer.
#: ...         _increment(_append_to_tuple),  # This is a transducer.
#: ...         lambda s, x: x + s,  # This is another transducer.
#: ...     ),
#: ...    lambda s, _: s,
#: ...    [(), 0],
#: ...    [1, 2, 3],
#: ... )
#: ((2, 3, 4), 6)
juxt = functional_generic.compose_left(
    operator.pack,
    enumerate,
    sync.map(sync.star(_transform_by_key(_replace_index))),
    sync.star(functional_generic.compose),
)


def map(f: Callable[[Any], Any]):
    """
    Transducer version of `map`.

    >>> transducer.transduce(
        transducer.map(lambda x: x + 1),
        lambda previous, current: (*previous, current),
        (),
        [1, 2, 3],
    )
    (3, 4, 5)
    """
    return lambda step: lambda s, current: step(s, f(current))


def filter(f: Callable[[Any], bool]):
    """
    Transducer version of `filter`.

    >>> transducer.transduce(
        transducer.filter(lambda x: x > 2),
        lambda previous, current: (*previous, current),
        [],
        [1, 2, 3]
    )
    (3,)
    """
    return lambda step: lambda s, x: step(s, x) if f(x) else s


def concat(step: Reducer):
    """Flattens collections using the `step` function into a single collection."""
    return lambda s, x: functools.reduce(step, x, s)


def mapcat(f: Callable[[Any], Iterable[Any]]):
    """Maps over a collection, then flattens the results into a single collection."""
    return functional_generic.compose(map(f), concat)


def groupby(key: Callable[[Any], Any], reducer: Reducer, initial):
    """Like `groupby_many`, just with a key function that returns a single element."""
    return groupby_many(
        functional_generic.compose_left(key, construct.wrap_tuple),
        reducer,
        initial,
    )


def groupby_many(keys: Callable[[Any], Iterable], reducer: Reducer, initial):
    """Given a `keys` function, that maps an element into multiple keys, transduces the collection into a dictionary of key to group of matching elements.

    >>> transducer.transduce(
        transducer.groupby_many(
            lambda x: ("even",) if x % 2 == 0 else ("odd",),
            lambda s, x: (*s, x),
            (),
        ),
        lambda s, _: s,
        {},
        [1, 2, 3, 4, 5],
    )
    {"even": (2, 4), "odd": (1, 3, 5)}
    """
    return functional_generic.compose(
        mapcat(
            functional_generic.compose_left(
                functional_generic.juxt(keys, construct.wrap_tuple),
                sync.star(itertools.product),
            ),
        ),
        lambda step: lambda s, x: step(
            dict_utils.add_key_value(x[0], reducer(s.get(x[0], initial), x[1]))(s),
            x,
        ),
    )


def count_by(keys: Callable[[Any], Iterable]):
    """Given a `keys` function, that maps an element into multiple keys, transduces the collection into a dictionary of key to count.

    >>> transducer.transduce(
        transducer.count_by(
            lambda x: ("even",) if x % 2 == 0 else ("odd",),
        ),
        lambda s, _: s,
        {},
        [1, 2, 3, 4, 5],
    )
    {"even": 2, "odd": 3}
    """
    return groupby_many(keys, lambda s, _: s + 1, 0)



import asyncio
import functools
import inspect
import itertools
from operator import not_
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Generator,
    Iterable,
    Mapping,
    Text,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from gamla import (
    apply_utils,
    construct,
    data,
    dict_utils,
    excepts_decorator,
    functional,
    operator,
)
from gamla.optimized import async_functions, sync


def compose_left(*funcs):
    """Compose sync and async functions to operate in series.

    Returns a function that applies other functions in sequence. The returned
    function will be an async function iff at least one of the functions in the
    sequence is async.

    Functions are applied from left to right so that
    ``compose_left(f, g, h)(x, y)`` is the same as ``h(g(f(x, y)))``.

    >>> inc = lambda i: i + 1
    >>> compose_left(inc, str)(3)
    '4'

    See Also:
        compose
        pipe
    """
    return compose(*reversed(funcs))


def curried_map(f):
    """Constructs a function that maps elements of a given iterable using the given function.

    Returns an async function iff `f` is async, else returns a sync function.

    >>> inc = lambda i: i + 1
    >>> curried_map(inc)([3, 4, 5])
    [4, 5, 6]
    """
    if asyncio.iscoroutinefunction(f):
        return async_functions.map(f)
    return sync.map(f)


def curried_to_binary(f):
    """Constructs a function from a given higher order function and returns its first order counterpart.
    The given higher order function, `f` is must be a unary function
    Returns an async function iff `f` is async, else returns a sync function.

    >>> inc = lambda i: i + 1
    >>> f = curried_to_binary(curried_map)
    >>> f(inc, [1, 2, 3])
    [2, 3, 4]
    """

    def internal(param1, param2):
        return f(param1)(param2)

    return internal


def any_is_async(funcs):
    return any(map(asyncio.iscoroutinefunction, funcs))


# Copying `toolz` convention.
# TODO(uri): Far from a perfect id, but should work most of the time.
# Improve by having higher order functions create meaningful names (e.g. `map`).
def _get_name_for_function_group(funcs):
    return "_OF_".join(map(lambda x: x.__name__, funcs))


def _match_return_typing(x: Callable, y: Callable) -> Dict:
    if hasattr(y, "__annotations__"):
        if "return" in y.__annotations__:
            return {**x.__annotations__, "return": y.__annotations__["return"]}
        if "return" in x.__annotations__:
            return dict_utils.remove_key("return")(x.__annotations__)
        return x.__annotations__
    if "return" in x.__annotations__:
        return dict_utils.remove_key("return")(x.__annotations__)
    return x.__annotations__


def compose(*funcs):
    """Compose sync and async functions to operate in series.

    Returns a function that applies other functions in sequence. The returned
    function will be an async function iff at least one of the functions in the
    sequence is async.

    Functions are applied from right to left so that
    ``compose(f, g, h)(x, y)`` is the same as ``f(g(h(x, y)))``.

    >>> inc = lambda i: i + 1
    >>> compose(str, inc)(3)
    '4'

    See Also:
        compose_left
        pipe
    """
    # TODO(uri): Enable after optimization checks.
    # prev = funcs[0]
    # for f in funcs[1:]:
    #     if not type_safety.composable(prev, f, None):
    #         logging.error(f"composing {prev} after {f}, which don't fit")
    #     prev = f

    if not funcs:
        return operator.identity

    if len(funcs) == 1:
        return funcs[0]

    if any_is_async(funcs):
        composed = async_functions.compose(*funcs)
    else:
        composed = sync.compose(*funcs)
    composed = functools.wraps(operator.last(funcs))(composed)
    frame = inspect.currentframe().f_back.f_back
    composed.__code__ = composed.__code__.replace(
        co_name=f"{frame.f_code.co_filename}:{frame.f_lineno}",
        co_filename=frame.f_code.co_filename,
        co_firstlineno=frame.f_lineno,
    )
    composed.__name__ = _get_name_for_function_group(funcs)
    composed.__annotations__ = _match_return_typing(composed, operator.head(funcs))
    return composed


def compose_many_to_one(incoming: Iterable[Callable], f: Callable):
    """Returns a function that applies an itterable of other functions into a
    single sink function. The returned function will be an async function iff
    at least one of the given functions is async.

    ``compose_many_to_one([f, g, k], h)(x, y)`` is the same as ``h(f(x,y), g(x, y), k(x, y))``.

    >>> compose_many_to_one([sum, sum], lambda x, y: x + y)([1, 2, 3])
    12

    See Also:
        juxt
        compose_left
    """
    return compose_left(juxt(*incoming), star(f))


def after(f1):
    """Second-order composition of `f1` over `f2`.

    A 'delayed' pipeline, i.e return a function that, given f2, will wait for f2's arguments, and when given, will return f1(f2(args)).
    >>> allmap = gamla.compose_left(gamla.map, gamla.after(all))
    >>> allmap(lambda x: x == 1)([1, 2, 3])
    False
    """

    def after(f2):
        return compose(f1, f2)

    return after


def before(f1):
    """Second-order composition of `f2` over `f1`."""

    def before(f2):
        return compose_left(f1, f2)

    return before


def lazyjuxt(
    *funcs: Tuple[Callable, ...]
) -> Union[Callable[..., Generator], Callable[..., Coroutine[None, None, Tuple]]]:
    """Create a function that applies each function in `funcs` to its arguments and returns a generator for the results.

    Applies the supplied functions lazily as the returned generator is iterated.
    Reverts to eager implementation if any of `funcs` is async.

    >>> inc = lambda x: x + 1
    >>> double = lambda x: x * 2
    >>> tuple(lazyjuxt(inc, double)(10))
    (11, 20)
    """
    if any_is_async(funcs):
        funcs = tuple(map(after(async_functions.to_awaitable), funcs))

        async def lazyjuxt_async(*args, **kwargs):
            return await asyncio.gather(*map(lambda f: f(*args, **kwargs), funcs))

        return lazyjuxt_async

    def lazyjuxt(*args, **kwargs):
        for f in funcs:
            yield f(*args, **kwargs)

    return lazyjuxt


def juxt(*funcs: Callable) -> Callable[..., Tuple]:
    """Create a function that applies each function in :funcs: to its arguments and returns a tuple of the results.

    >>> inc = lambda x: x + 1
    >>> double = lambda x: x * 2
    >>> juxt(inc, double)(10)
    (11, 20)
    """
    if any_is_async(funcs):
        funcs = tuple(map(after(async_functions.to_awaitable), funcs))

        async def juxt_async(*args, **kwargs):
            return await asyncio.gather(*map(lambda f: f(*args, **kwargs), funcs))

        return compose(tuple, juxt_async)

    def juxt(*args, **kwargs):
        return tuple(func(*args, **kwargs) for func in funcs)

    return juxt


#: Pass a value through a list of functions, return `True` iff all functions returned `True`-ish values.
#:
#: >>> f = alljuxt(gamla.identity, gamla.greater_than(1), gamla.greater_than(10))
#: >>> f(100)
#: True
#: >>> f(10)
#: False
alljuxt = compose(after(all), lazyjuxt)

#: Pass a value through a list of functions, return `True` if at least one function returned a `True`-ish value.
#   Note: evaluation is lazy, i.e. returns on first `True`.
#:
#: >>> f = anyjuxt(gamla.identity, gamla.greater_than(1), gamla.greater_than(10))
#: >>> f(100)
#: True
#: >>> f(10)
#: True
#: >>> f(0)
#: False
anyjuxt = compose(after(any), lazyjuxt)

#: Create a function that calls the supplied functions, and chains the results.
#: Assumes the supplied functions return Iterables.
#: >>> f = juxtcat(range,range)
#: >>> tuple(f(5))
#: (0 ,1 ,2 ,3 ,4 ,0 ,1 ,2 ,3 ,4)
juxtcat = compose(after(itertools.chain.from_iterable), lazyjuxt)


def ternary(condition, f_true, f_false):
    """Returns a function that computes `f_true` or `f_false` according to
    `condition`. The functions are applied with the same input.
    The returned function will be an async function if at least one of the given
    functions is async.

    >>> f = ternary(gamla.greater_than(5), gamla.identity, lambda i: -i)
    >>> f(6)
    '6'
    >>> f(3)
    '-3'
    """
    if any_is_async([condition, f_true, f_false]):

        async def ternary_inner_async(*args, **kwargs):
            return (
                await async_functions.to_awaitable(f_true(*args, **kwargs))
                if await async_functions.to_awaitable(condition(*args, **kwargs))
                else await async_functions.to_awaitable(f_false(*args, **kwargs))
            )

        return ternary_inner_async

    return sync.ternary(condition, f_true, f_false)


def when(condition: Callable, f_true: Callable) -> Callable:
    """Returns `f_true(args)` if `condition(args)` returns true, else returns args.

    >>> f = when(gamla.greater_than(5), lambda i: -i)
    >>> f(6)
    '-6'
    >>> f(3)
    '3'
    """
    return ternary(condition, f_true, operator.identity)


def unless(condition, f_false):
    """Returns a function that computes `f_false` if `condition` is met.
    Otherwise will return the input unchanged.
    `condition` and `f_false` are applied with the same input.


    >>> f = unless(gamla.greater_than(5), lambda i: -i)
    >>> f(6)
    '6'
    >>> f(3)
    '-3'
    """
    return ternary(condition, operator.identity, f_false)


def first(*funcs, exception_type: Type[Exception]):
    """Constructs a function that computes all functions from `funcs`, and returns the first function that doesn't throw an exception of type `exception_type`. The
    function is async if at least one of the given functions is async. If all functions throw the
    given `exception_type`, `exception_type` will be raised.

    >>> f = gamla.first(gamla.second, gamla.head, exception_type=StopIteration)
    >>> f([1,2])
    '2'
    >>> f([1])
    '1'
    >>> f([])
    StopIteration raised
    """
    if any_is_async([*funcs]):

        async def inner_async(*args, **kwargs):
            for func in funcs:
                try:
                    return await async_functions.to_awaitable(func(*args, **kwargs))
                except exception_type:
                    pass
            raise exception_type

        return inner_async

    def inner(*args, **kwargs):
        for func in funcs:
            try:
                return func(*args, **kwargs)
            except exception_type:
                pass
        raise exception_type

    return inner


class PipeNotGivenAnyFunctions(Exception):  # noqa
    pass


def pipe(val, *funcs):
    """Pipe a value through a sequence of functions

    I.e. ``pipe(val, f, g, h)`` is equivalent to ``h(g(f(val)))``

    >>> double = lambda i: 2 * i
    >>> pipe(3, double, str)
    '6'
    """
    if not funcs:
        raise PipeNotGivenAnyFunctions
    if any_is_async(funcs):
        return async_functions.compose(*reversed(funcs))(val)
    for f in funcs:
        val = f(val)
    return val


#: Map an iterable using a function, return `True` iff all mapped values are `True`-ish.
#: >>> f = allmap(lambda x: x % 2 == 0)
#: >>> f([1, 2, 3, 4, 5])
#: False
#: >>> f([2, 4, 6, 8, 10])
#: True
allmap = compose(after(all), curried_map)

#: Map an iterable using a function, return `True` if at least one mapped value is `True`-ish.
#: Note: evaluation is lazy, i.e. returns on first `True`.
#: >>> f = anymap(lambda x: x % 2 == 0)
#: >>> f([1, 2, 3, 4, 5])
#: True
#: >>> f([1, 3, 5, 7, 9])
#: False
anymap = compose(after(any), curried_map)

#: Constructs a function that applies the given function to items of a given dictionary.
#: Returns an async function iff the filter function is async, else returns a sync function.
#:
#: >>> f = itemmap(gamla.star(lambda key, val: (key, val + key)))
#: >>> f({1: 2, 2: 3})
#: {1: 3, 2: 5}
itemmap = compose(after(dict), before(dict.items), curried_map)

#: Creates a function that maps supplied mapper over the keys of a dict.
#:
#: >>> f = keymap(lambda k: k + 1)
#: >>> f({1:"a",2:"b",3:"c",4:"d"})
#: {
#:   2: "a",
#:   3: "b",
#:   4: "c",
#:   5: "d"
#: }
keymap = compose(
    itemmap,
    lambda f: juxt(f, operator.second),
    before(operator.head),
)

#: Creates a function then maps the supplied mapper over the values of a dict.
#:
#: >>> f = valmap(gamla.add(1))
#: >>> f({ "a": 1, "b": 2, "c": 3 })
#: {
#: "a": 2
#: "b": 3
#: "c": 4
#: }
valmap = compose(
    itemmap,
    lambda f: juxt(operator.head, f),
    before(operator.second),
)


def pair_with(f):
    """Returns a function that given a value x, returns a tuple of the form: (f(x), x).

    >>> add_one = pair_with(lambda x: x + 1)
    >>> add_one(3)
    (4, 3)
    """
    return juxt(f, operator.identity)


def pair_right(f):
    """Returns a function that given a value x, returns a tuple of the form: (x, f(x)).

    >>> add_one = pair_right(lambda x: x + 1)
    >>> add_one(3)
    (3, 4)
    """
    return juxt(operator.identity, f)


#: Constructs a function that filters elements of a given iterable for which function returns true.
#: Returns an async function iff the filter function is async, else returns a sync function.
#:
#: >>> f = curried_filter(gamla.greater_than(10))
#: >>> f([1, 2, 3, 11, 12, 13])
#: [11, 12, 13]
curried_filter = compose(
    after(
        sync.compose(
            sync.map(operator.second),
            sync.filter(operator.head),
        ),
    ),
    curried_map,
    pair_with,
)

#: Constructs a function that filters items of a given dictionary for which function returns true.
#: Returns an async function iff the filter function is async, else returns a sync function.
#:
#: >>> f = itemfilter(
#: ...     alljuxt(
#: ...         compose_left(gamla.head, gamla.contains("gamla")),
#: ...         compose_left(gamla.second, gamla.greater_than(10)),
#: ...     )
#: ... )
#: >>> f({"gamla": 11, "gaml": 9, "f":12})
#: {'gamla': 11}
itemfilter = compose(after(dict), before(dict.items), curried_filter)

#: Create a function that filters a dict using a predicate over keys.
#:
#: >>> f = keyfilter(lambda k: k > 2)
#: >>> f({1:"a",2:"b",3:"c",4:"d"})
#: {
#:   3: "c",
#:   4: "d"
#: }
keyfilter = compose(
    itemfilter,
    before(operator.head),
)

#: Create a function that filters a dict using a predicate over values.
#:
#: >>> f = valefilter(lambda k: k > 2)
#: >>> f({"a": 1, "b": 2, "c": 3, "d": 4})
#: {
#:   "c": 3,
#:   "d": 4
#: }
valfilter = compose(
    itemfilter,
    before(operator.second),
)

#: Complement of a boolean function.
#:
#: >>> f = complement(gamla.greater_than(5))
#: >>> f(10)
#: False
#: >>> f(1)
#: True
complement = after(not_)

#: Constructs a function that removes elements of a given iterable for which function returns true.
#: Returns an async function iff the filter function is async, else returns a sync function.
#:
#: >>> f = remove(gamla.greater_than(10))
#: >>> tuple(f([1, 2, 3, 11, 12, 13]))
#: (1, 2, 3)
remove = compose(curried_filter, complement)

_make_async = sync.after(async_functions.to_awaitable)


def case(predicates_and_mappers: Tuple[Tuple[Callable, Callable], ...]):
    """Applies mappers to values according to predicates. If no predicate matches, raises `gamla.functional_generic.NoConditionMatched`.
    >>> f = case(((gamla.less_than(10), gamla.identity), (gamla.greater_than(10), gamla.add(100))))
    >>> f(5)
    5
    >>> f(15)
    115
    >>> f(10)
    `NoConditionMatched`
    """
    if any_is_async(operator.concat(predicates_and_mappers)):
        predicates, mappers = zip(*predicates_and_mappers)
        predicates = tuple(map(_make_async, predicates))
        mappers = tuple(map(_make_async, mappers))

        async def case_async(*args, **kwargs):
            for is_matched, mapper in zip(
                await asyncio.gather(*map(lambda f: f(*args, **kwargs), predicates)),
                mappers,
            ):
                if is_matched:
                    return await mapper(*args, *kwargs)
            raise sync.NoConditionMatched({"input args": args, "input kwargs": kwargs})

        return case_async

    return sync.case(predicates_and_mappers)


#: Applies functions to values according to predicates given in a dict. Raises `gamla.functional_generic.NoConditionMatched` if no predicate matches.
#: >>> f = case_dict({gamla.less_than(10): gamla.identity, gamla.greater_than(10): gamla.add(100)})
#: >>> f(5)
#: 5
#: >>> f(15)
#: 115
#: >>> f(10)
#: `NoConditionMatched`
case_dict = sync.compose_left(dict.items, tuple, case)


async def _await_dict(value):
    if isinstance(value, dict) or isinstance(value, data.frozendict):
        return await pipe(
            value,
            # In case input is a `frozendict`.
            dict,
            valmap(_await_dict),
        )
    if isinstance(value, Iterable):
        return await pipe(value, async_functions.map(_await_dict), type(value))
    return await async_functions.to_awaitable(value)


def map_dict(nonterminal_mapper: Callable, terminal_mapper: Callable):
    f = sync.map_dict(nonterminal_mapper, terminal_mapper)
    if any_is_async([nonterminal_mapper, terminal_mapper]):
        return compose_left(f, _await_dict)
    return f


def _iterdict(d):
    results = []
    map_dict(operator.identity, results.append)(d)
    return results


_has_coroutines = compose_left(_iterdict, any_is_async)


def apply_spec(spec: Dict):
    """Named transformations of a value using named functions.

    >>> spec = {"len": len, "sum": sum}
    >>> apply_spec(spec)([1,2,3,4,5])
    {'len': 5, 'sum': 15}

    Notes:
    - The dictionary can be nested.
    - Returned function will be async iff any leaf is an async function.
    """
    if _has_coroutines(spec):

        async def apply_spec_async(*args, **kwargs):
            return await map_dict(
                operator.identity,
                compose_left(
                    apply_utils.apply(*args, **kwargs),
                    async_functions.to_awaitable,
                ),
            )(spec)

        return apply_spec_async

    def apply_spec_sync(*args, **kwargs):
        return sync.map_dict(operator.identity, lambda f: f(*args, **kwargs))(spec)

    return apply_spec_sync


#: Construct a function that applies the i'th function in an iterable
#  on the i'th element of a given iterable
#:
#: Note: Number of functions should be equal to the number of elements in the given iterable
#:
#: >>> stack([lambda x:x+1, lambda x:x-1])((5, 5))
#: (6, 4)
stack = compose_left(
    enumerate,
    sync.map(
        sync.star(lambda i, f: compose(f, lambda x: x[i])),
    ),
    sync.star(juxt),
)


def bifurcate(*funcs):
    """Serially run each function on tee'd copies of a sequence.
    If the sequence is a generator, it is duplicated so it will not be exhausted (which may incur a substantial memory signature in some cases).
    >>> f = bifurcate(sum, gamla.count)
    >>> seq = map(gamla.identity, [1, 2, 3, 4, 5])
    >>> f(seq)
    (15, 5)
    """
    return compose_left(iter, lambda it: itertools.tee(it, len(funcs)), stack(funcs))


def value_to_dict(key: Text):
    """Converts a string and Any input to a dict object.

    >>> value_to_dict("hello")("world")
    {'hello': 'world'}
    """
    return compose_left(
        construct.wrap_tuple,
        functional.prefix(key),
        construct.wrap_tuple,
        dict,
    )


_ReducerState = TypeVar("_ReducerState")
_ReducedElement = TypeVar("_ReducedElement")
Reducer = Callable[[_ReducerState, _ReducedElement], _ReducerState]


def reduce_curried(
    reducer: Reducer,
    initial_value: _ReducerState,
) -> Callable[[Iterable[_ReducedElement]], _ReducerState]:
    if asyncio.iscoroutinefunction(reducer):

        async def reduce_async(elements):
            state = initial_value
            for element in elements:
                state = await reducer(state, element)
            return state

        return reduce_async

    return sync.reduce(reducer, initial_value)


def scan(
    reducer: Reducer,
    initial_value: _ReducerState,
) -> Callable[[Iterable[_ReducedElement]], Tuple[_ReducerState, ...]]:
    """Like `reduce`, but keeps history of states.

    See https://en.wikipedia.org/wiki/Prefix_sum#Scan_higher_order_function."""

    if asyncio.iscoroutinefunction(reducer):

        async def reduce_keeping_history_async(
            past_states: Tuple[_ReducerState, ...],
            element: _ReducedElement,
        ) -> Tuple[_ReducerState, ...]:
            return (*past_states, await reducer(operator.last(past_states), element))

        return reduce_curried(reduce_keeping_history_async, (initial_value,))

    def reduce_keeping_history(
        past_states: Tuple[_ReducerState, ...],
        element: _ReducedElement,
    ) -> Tuple[_ReducerState, ...]:
        return (*past_states, reducer(operator.last(past_states), element))

    return reduce_curried(reduce_keeping_history, (initial_value,))


#: Constructs a function that will return the first element of an iterable,
#: that returns True when used with the the given function. If no element
#: in the iterable returns True, None is returned.
#:
#: >>> f = find(gamla.greater_than(10))
#: >>> f([1, 2, 3, 11, 12, 13])
#: 11
#: >>> f([1, 2, 3])
#: None
find = compose(
    after(
        excepts_decorator.excepts(
            StopIteration,
            construct.just(None),
            operator.head,
        ),
    ),
    curried_filter,
)


#: Constructs a function that will return the index of an element of an iterable,
#: that returns True when used with the the given function. If no element
#  in the iterable returns True, -1 is returned.
#:
#: >>> f = find(gamla.greater_than(10))
#: >>> f([1, 2, 3, 11, 12, 13])
#: 11
#: >>> f([1, 2, 3])
#: -1
#:
#: See Also:
#:  - find
find_index = compose_left(
    before(operator.second),
    find,
    before(enumerate),
    after(ternary(operator.equals(None), construct.just(-1), operator.head)),
)


def _inner_merge_with(dicts):
    if len(dicts) == 1 and not isinstance(dicts[0], Mapping):
        dicts = dicts[0]
    result = {}
    for d in dicts:
        for k, v in d.items():
            if k in result:
                result[k].append(v)
            else:
                result[k] = [v]
    return result


map_filter_empty = compose_left(curried_map, after(curried_filter(operator.identity)))


def merge_with(f):
    if asyncio.iscoroutinefunction(f):

        async def merge_with(*dicts):
            result = _inner_merge_with(dicts)
            return await valmap(f)(result)

        return merge_with

    def merge_with(*dicts):
        result = _inner_merge_with(dicts)
        return sync.valmap(f)(result)

    return merge_with


merge = sync.compose_left(lambda *x: x if len(x) > 1 else x[0], sync.merge)
mapcat = compose_left(curried_map, after(operator.concat))

_K = TypeVar("_K")


def _groupby_helper(state, current):
    x = state or []
    x.append(current)
    return x


def groupby(
    key: Callable[[_ReducedElement], _K],
) -> Callable[[Iterable[_ReducedElement]], Mapping[_K, Tuple[_ReducedElement, ...]]]:
    """Return a mapping `{y: {x s.t. key(x) = y}}.`

    >>> names = ['alice', 'bob', 'barbara', 'frank', 'fred']
    >>> f = groupby(gamla.head)
    >>> f(names)
    {"a": ("alice",),
     "b": ("bob", "barbara"),
     "f": ("frank", "fred")}
    """
    return compose_left(
        functional.groupby_many_reduce(
            compose_left(key, construct.wrap_tuple),
            _groupby_helper,
        ),
        sync.valmap(tuple),
    )


def is_generator(iterable):
    return hasattr(iterable, "__iter__") and not hasattr(iterable, "__len__")


def side_effect(f: Callable):
    """Runs `f` on `x`, returns `x`

    >>> log_and_add = compose_left(side_effect(print), add(1)))
    >>> log_and_add(2)
    2
    3
    """
    if asyncio.iscoroutinefunction(f):

        async def do(x):
            if is_generator(x):
                a, b = itertools.tee(x)
                await f(a)
                return b
            await f(x)
            return x

    else:

        def do(x):
            if is_generator(x):
                a, b = itertools.tee(x)
                f(a)
                return b
            f(x)
            return x

    return do


def count_by_many(f: Callable[[Any], Iterable]) -> Dict[Any, int]:
    """Counts elements of a collection by a key function `f`.

    >>> count_by_many(
    ...     gamla.juxt(
    ...         operator.head,
    ...         gamla.last,
    ...     )
    ... )(["aa", "ab", "ac", "bc"])
    {'a': 3, 'b': 2, 'c': 2}
    """
    return functional.groupby_many_reduce(
        f,
        lambda x, _: x + 1 if x else 1,
    )


countby_many = count_by_many

#: Like `count_by_many` but with a function that returns a single key.
count_by = compose_left(after(construct.wrap_tuple), count_by_many)

#: Like `stack` but doesn't require additional brackets.
packstack = compose_left(operator.pack, stack)
#: Runs `packstack` with given functions, then runs `all` on the output.
allstack = compose_left(packstack, after(all))
#: Runs `packstack` with given functions, then runs `any` on the output.
anystack = compose_left(packstack, after(any))


def _choose_by_async(f_sync, f_async):
    def choose_by_async(f):
        if inspect.iscoroutinefunction(f):
            return f_async(f)
        return f_sync(f)

    return choose_by_async


#: Turns a variadic function into an unary one that gets a tuple of args to the original function.
#:
#: >>> pipe((2, 3), star(lambda x, y: x + y))
#: 5
star = _choose_by_async(sync.star, async_functions.star)

#: Turns a variadic function into an unary one that gets a dict of keywoded args to the original function.
#:
#: >>> pipe(({"x": 2, "y": 3}), double_star(lambda x, y: x + y))
#: 5
double_star = _choose_by_async(sync.double_star, async_functions.double_star)

#: Return a dict with number of occurrences of each value in a sequence.
#: >>> frequencies(['cat', 'cat', 'ox', 'pig', 'pig', 'cat'])
#: {'cat': 3, 'ox': 1, 'pig': 2}
frequencies = count_by(operator.identity)



import dataclasses
from typing import Any, Callable

from gamla import construct, currying, dict_utils, functional_generic, operator
from gamla.optimized import async_functions, sync


@currying.curry
def tree_reduce(get_children: Callable, reduce_fn: Callable, tree_node):
    """Reduces a tree from the bottom up.

    Given `get_children`, a function from a node to its children,
    and `reduce_fn`, which gets a node and the results of the reduce on the children,
    reduces a tree upwards.
    """
    return reduce_fn(
        tree_node,
        map(tree_reduce(get_children, reduce_fn), get_children(tree_node)),
    )


def tree_reduce_async(get_children: Callable, reduce_fn: Callable):
    """Async version of `tree_reduce`."""

    async def tree_reduce_async_inner(tree_node):
        return await reduce_fn(
            tree_node,
            await functional_generic.pipe(
                tree_node,
                get_children,
                functional_generic.curried_map(
                    tree_reduce_async(get_children, reduce_fn),
                ),
            ),
        )

    return tree_reduce_async_inner


def map_reduce_tree(children: Callable, reducer: Callable, mapper: Callable):
    """Like `tree_reduce`, but allows for an async map stage first, so it can be parallelized."""
    return functional_generic.compose_left(
        functional_generic.juxt(
            functional_generic.compose_left(
                children,
                functional_generic.curried_map(
                    (
                        async_functions.thunk
                        if functional_generic.any_is_async([children, reducer, mapper])
                        else sync.thunk
                    )(map_reduce_tree, children, reducer, mapper),
                ),
                tuple,
            ),
            mapper,
        ),
        functional_generic.star(reducer),
    )


@dataclasses.dataclass(frozen=True)
class KeyValue:
    key: Any
    value: Any


_is_terminal = sync.anyjuxt(
    operator.is_instance(str),
    operator.is_instance(int),
    operator.is_instance(float),
)


def _get_children(element):
    return functional_generic.case_dict(
        {
            _is_terminal: construct.just(()),
            operator.is_instance(tuple): operator.identity,
            operator.is_instance(list): operator.identity,
            operator.is_instance(dict): functional_generic.compose_left(
                dict.items,
                sync.map(sync.star(KeyValue)),
            ),
            operator.is_instance(KeyValue): functional_generic.compose_left(
                lambda x: x.value,
                sync.ternary(
                    _is_terminal,
                    construct.wrap_tuple,
                    _get_children,
                ),
            ),
        },
    )(element)


_MATCHED = "matched"
_UNMATCHED = "unmatched"
_get_matched = dict_utils.itemgetter(_MATCHED)
_get_unmatched = dict_utils.itemgetter(_UNMATCHED)


def _make_matched_unmatched(matched, unmatched):
    return {_MATCHED: matched, _UNMATCHED: unmatched}


_merge_children_as_matched = functional_generic.compose_left(
    sync.mapcat(sync.juxtcat(_get_matched, _get_unmatched)),
    tuple,
    functional_generic.pair_right(construct.just(())),
    sync.star(_make_matched_unmatched),
)

_merge_children = sync.compose_left(
    functional_generic.bifurcate(
        sync.compose_left(sync.mapcat(_get_matched), tuple),
        sync.compose_left(
            sync.mapcat(_get_unmatched),
            tuple,
        ),
    ),
    sync.star(_make_matched_unmatched),
)


@currying.curry
def _get_anywhere_reducer(predicate: Callable, node, children):
    if _is_terminal(node):
        return _make_matched_unmatched((), (node,))
    if isinstance(node, KeyValue) and predicate(node.key):
        return _merge_children_as_matched(children)
    return _merge_children(children)


def get_leaves_by_ancestor_predicate(predicate: Callable):
    """Gets a predicate, and builds a function that gets a dictionary, potentially nested and returns an iterable of leaf values.
    The values returned are of leafs where some ancestor (possibly indirect) passes the predicate.

    >>> gamla.pipe({"x": {"y": (1, 2, 3)}}, gamla.get_leaves_by_ancestor_predicate(gamla.equals("x")), tuple)
    (1, 2, 3)
    >>> gamla.pipe({"x": {"y": (1, 2, 3)}}, gamla.get_leaves_by_ancestor_predicate(gamla.equals("z")), tuple)
    ()

    Useful for retrieving values from large json objects, where the exact path is unimportant.
    """
    return functional_generic.compose_left(
        tree_reduce(_get_children, _get_anywhere_reducer(predicate)),
        _get_matched,
    )


@currying.curry
def _filter_leaves_reducer(predicate, node, children):
    if _is_terminal(node) and predicate(node):
        return (node,)
    return operator.concat(children)


def filter_leaves(predicate: Callable):
    """Gets a predicate, and builds a function that gets a dictionary, potentially nested and returns an iterable of leaf values.
    The values returned are of leafs that pass the predicate.

    >>> gamla.pipe({"x": {"y": (1, 2, 3)}}, gamla.filter_leaves(gamla.greater_than(2)), tuple)
    (3,)

    Useful for retrieving values from large json objects, where the exact path is unimportant.
    """
    return tree_reduce(_get_children, _filter_leaves_reducer(predicate))


#: Reduce a JSON like tree.
json_tree_reduce = tree_reduce(_get_children)



from gamla import higher_order


def test_prepare_and_apply():
    def increment(x):
        return x + 1

    def decrement(x):
        return x - 1

    def conditional_transformation(x):
        return increment if x < 10 else decrement

    assert higher_order.prepare_and_apply(conditional_transformation)(15) == 14


async def test_prepare_and_apply_async():
    async def increment(x):
        return x + 1

    async def decrement(x):
        return x - 1

    def conditional_transformation(x):
        return increment if x < 10 else decrement

    assert (
        await higher_order.prepare_and_apply_async(conditional_transformation)(15) == 14
    )



from gamla import async_functions, dict_utils, functional_generic, graph_async, operator


def _reduce_max(children, current):
    return max(children + (current,))


_get_neighbors = dict_utils.dict_to_getter_with_default(
    (),
    {1: (1, 2, 3, 5), 2: (4,), 3: (1, 2)},
)


async def test_reduce_graph_async1():
    set_instance = set()
    assert (
        await graph_async.reduce_graph_async(
            _reduce_max,
            functional_generic.compose_left(
                _get_neighbors,
                async_functions.to_awaitable,
            ),
            set_instance.add,
            operator.contains(set_instance),
            1,
        )
        == 5
    )


async def test_reduce_graph_async2():
    set_instance = set()
    assert (
        await graph_async.reduce_graph_async(
            lambda children, current: sum(children) + current,
            functional_generic.compose_left(
                dict_utils.dict_to_getter_with_default(
                    (),
                    {1: (1, 2, 3, 5), 2: (4,), 3: (1, 2)},
                ),
                async_functions.to_awaitable,
            ),
            set_instance.add,
            operator.contains(set_instance),
            1,
        )
        == 15
    )


async def test_reduce_graph_async_reducer():
    set_instance = set()
    assert (
        await graph_async.reduce_graph_async(
            functional_generic.compose_left(
                _reduce_max,
                async_functions.to_awaitable,
            ),
            functional_generic.compose_left(
                _get_neighbors,
                async_functions.to_awaitable,
            ),
            set_instance.add,
            operator.contains(set_instance),
            1,
        )
        == 5
    )



import functools
from operator import getitem
from typing import Any, Callable, Dict, Iterable

from gamla import construct, currying, functional, functional_generic, operator
from gamla.optimized import sync


def itemgetter(attr):
    """Retrieves a dictionary entry by its key `attr`.

    >>> itemgetter("a")({"a": 1})
    1
    """

    def itemgetter(obj):
        return obj[attr]

    return itemgetter


def itemgetter_or_none(attr):
    """Retrieves a dictionary entry by its key `attr`. Returns `None` if the key is not there.

    >>> itemgetter_or_none("a")({"a": 1})
    1

    >>> itemgetter_or_none("b")({"a": 1})
    None
    """

    def itemgetter_or_none(obj):
        return obj.get(attr, None)

    return itemgetter_or_none


def itemgetter_with_default(default, attr):
    """Retrieves a dictionary entry by its key `attr`. Returns `default` if the key is not there.

    >>> itemgetter_with_default(0, "a")({"a": 1})
    1
    >>> itemgetter_with_default(0, "b")({"a": 1})
    0
    """

    def itemgetter_with_default(d):
        try:
            return d[attr]
        except (KeyError, IndexError):
            return default

    return itemgetter_with_default


@currying.curry
def get_or_transform(f, d):
    def get_or_transform(x):
        if x in d:
            return d[x]
        return f(x)

    return get_or_transform


get_or_identity = get_or_transform(operator.identity)


def get_in(keys: Iterable):

    """Creates a function that returns coll[i0][i1]...[iX] where [i0, i1, ..., iX]==keys.

    >>> get_in(["a", "b", 1])({"a": {"b": [0, 1, 2]}})
    1
    """

    def get_in(coll):
        return functools.reduce(getitem, keys, coll)

    return get_in


def get_in_with_default(keys: Iterable, default):
    """`get_in` function, returning `default` if a key is not there.

    >>> get_in_with_default(["a", "b", 1], 0)({"a": {"b": [0, 1, 2]}})
    1
    >>> get_in_with_default(["a", "c", 1], 0)({"a": {"b": [0, 1, 2]}})
    0
    """
    getter = get_in(keys)

    def get_in_with_default(x):
        try:
            return getter(x)
        except (KeyError, IndexError, TypeError):
            return default

    return get_in_with_default


def get_in_or_none(keys: Iterable):
    """`get_in` function, returning `None` if a key is not there.

    >>> get_in_or_none(["a", "b", 1])({"a": {"b": [0, 1, 2]}})
    1
    >>> get_in_or_none(["a", "c", 1])({"a": {"b": [0, 1, 2]}})
    None
    """
    return get_in_with_default(keys, None)


def get_in_or_none_uncurried(keys: Iterable, coll):
    """Returns coll[i0][i1]...[iX] where [i0, i1, ..., iX]==keys and `None` if a key is not there.

    >>> get_in_or_none_uncurried(["a", "b", 1],{"a": {"b": [0, 1, 2]}})
    1
    >>> get_in_or_none_uncurried(["a", "c", 1], {"a": {"b": [0, 1, 2]}})
    None
    """
    return get_in_or_none(keys)(coll)


@currying.curry
def dict_to_getter_with_default(default, d: Dict):
    """Turns a dictionary into a function from key to value or default if key is not there.

    >>> dict_to_getter_with_default(None, {1:1})(1)
    1
    >>> dict_to_getter_with_default(None, {1:1})(2)
    None
    """

    def dict_to_getter_with_default(key):
        return d.get(key, default)

    return dict_to_getter_with_default


def _return_after_n_calls(n, value):
    if n == 0:
        return value

    def return_after_n_calls(_):
        return _return_after_n_calls(n - 1, value)

    return return_after_n_calls


def make_index(
    steps: Iterable[Callable[[Iterable], Dict]],
) -> Callable[[Iterable], Any]:
    """Builds an index with arbitrary amount of steps from an iterable.

    >>> index = dict_utils.make_index(map(gamla.groupby, [gamla.head, gamla.second]))(["uri", "dani"])
    >>> index("d")("a")
    frozenset(["dani"])
    """
    steps = tuple(steps)
    if not steps:
        return frozenset
    return functional_generic.compose_left(
        operator.head(steps),
        functional_generic.valmap(make_index(steps[1:])),
        lambda d: lambda x: d.get(
            x,
            _return_after_n_calls(len(steps) - 1, frozenset()),
        ),
    )


def add_key_value(key, value):
    """Associate a key-value pair to the input dict.

    >>> add_key_value("1", "1")({"2": "2"})
    {'2': '2', '1': '1'}
    """

    def add_key_value(d):
        return functional.assoc_in(d, [key], value)

    return add_key_value


def remove_key(key):
    """Given a dictionary, return a new dictionary with 'key' removed.
    >>> remove_key("two")({"one": 1, "two": 2, "three": 3})
    {'one': 1, 'three': 3}
    """

    def remove_key(d: dict):
        updated = d.copy()
        del updated[key]
        return updated

    return remove_key


def rename_key(old: str, new: str) -> Callable[[dict], dict]:
    """Rename a key in a dictionary.

    >>> my_dict = {"name": "Danny", "age": 20}
    >>> rename_key("name", "first_name")(my_dict)
    {'first_name': 'Danny', 'age': 20}
    """
    return sync.keymap(sync.when(operator.equals(old), construct.just(new)))


def transform_item(key, f: Callable) -> Callable[[dict], dict]:
    """transform a value of `key` in a dict. i.e given a dict `d`, return a new dictionary `e` s.t e[key] = f(d[key]).

    >>> my_dict = {"name": "Danny", "age": 20}
    >>> transform_item("name", str.upper)(my_dict)
    {'name': 'DANNY', 'age': 20}
    """
    return functional_generic.itemmap(
        functional_generic.when(
            functional_generic.compose_left(operator.head, operator.equals(key)),
            functional_generic.packstack(operator.identity, f),
        ),
    )



import dataclasses
import functools
import hashlib
import heapq
import inspect
import itertools
import json
import os
import random
from concurrent import futures
from operator import truediv
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Sequence,
    Text,
    Tuple,
    TypeVar,
)

import heapq_max
import immutables
import toolz

from gamla import construct, currying, excepts_decorator, operator
from gamla.optimized import sync


def sort_by(key: Callable):
    """Return a new list containing all items from the iterable in ascending order, sorted by a key.
    >>> sort_by(len)(["hi!", "my", "name", "is"])
    ['my', 'is', 'hi!', 'name']
    """

    def sort_by(seq: Iterable):
        return sorted(seq, key=key)

    return sort_by


def sort_by_reversed(key: Callable):
    """Return a new list containing all items from the iterable in descending order, sorted by a key.
    >>> sort_by_reversed(lambda x: x % 10)([2231, 47, 19, 100])
    [19, 47, 2231, 100]
    """

    def sort_by_reversed(seq: Iterable):
        return sorted(seq, key=key, reverse=True)

    return sort_by_reversed


#: Return a new list containing all items from the iterable in ascending order
#: >>> sort([5,2,4,1])
#: '[1,2,4,5]'
sort = sort_by(operator.identity)

#: Return a new list containing all items from the iterable in descending order
#: >>> sort([5,2,4,1])
#: '[5,4,2,1]'
sort_reversed = sort_by_reversed(operator.identity)


def do_if(condition, fun):
    def inner_do_if(x):
        if condition(x):
            fun(x)
        return x

    return inner_do_if


def singleize(func: Callable) -> Callable:
    def wrapped(some_input):
        if isinstance(some_input, tuple):
            return func(some_input)
        return operator.head(func((some_input,)))

    async def wrapped_async(some_input):
        if isinstance(some_input, tuple):
            return await func(some_input)
        return operator.head(await func((some_input,)))

    if inspect.iscoroutinefunction(func):
        return wrapped_async
    return wrapped


def ignore_input(inner: Callable[[], Any]) -> Callable:
    """Returns `inner` function ignoring the provided inputs.

    >>> ignore_input(lambda: 0)(1)
    0
    """

    def ignore_and_run(*args, **kwargs):
        return inner()

    async def ignore_and_run_async(*args, **kwargs):
        return await inner()

    if inspect.iscoroutinefunction(inner):
        return ignore_and_run_async
    return ignore_and_run


def just_raise(exception):
    """Raises the given exception.

    >>> just_raise(KeyError)
    raise exception KeyError
    """
    raise exception


def make_raise(exception):
    """Returns a function that ignores the input and just raises the given exception.

    >>> f = make_raise(KeyError)
    >>> f(3)
    raise exception KeyError
    """

    def inner():
        raise exception

    return ignore_input(inner)


@currying.curry
def translate_exception(func: Callable, exc1: Exception, exc2: Exception):
    """A functional try/except block: if `func` fails with `exc1`, raise `exc2`.

    >>> from gamla import functional_generic
    >>> functional_generic.pipe(iter([]), translate_exception(next, StopIteration, ValueError))
    ValueError
    Note: `func` is assumed to be unary."""
    return excepts_decorator.excepts(exc1, make_raise(exc2), func)


def to_json(obj):
    """Return a `JSON` representation of a 'dictionary' or an object.

    >>> to_json({"one": 1, "two": 2})
    '{"one": 1, "two": 2}'
    """
    if hasattr(obj, "to_json"):
        return obj.to_json()
    return json.dumps(obj)


def compute_stable_json_hash(item) -> Text:
    """Only works on json valid data types."""
    return hashlib.sha1(
        json.dumps(
            item,
            sort_keys=True,
            separators=(",", ":"),
        ).encode("utf-8"),
    ).hexdigest()


@currying.curry
def assert_that_with_message(input_to_message: Callable, f: Callable):
    """Assert a function `f` on the input, printing the output of `input_to_message(input)` if assertion is False.

    >>> assert_that_with_message(just("Input is not 2!"), equals(2))(2)
    2
    >>> assert_that_with_message(just("Input is not 2!"), equals(2))(3)
    "Output is not 2!"
    """

    def assert_that_f(inp):
        assert f(inp), input_to_message(inp)
        return inp

    return assert_that_f


assert_that = assert_that_with_message(construct.just(""))


@currying.curry
def pmap(f, n_workers, it):
    # The `tuple` is for callers convenience (even without it, the pool is eager).
    return tuple(futures.ThreadPoolExecutor(max_workers=n_workers).map(f, it))


# To get a unique caching key for each function invocation, we take `args` and `items()`
# of `kwargs` and sort them (by keys), while also marking the beginning of `kwargs`.
# Inspired by: http://code.activestate.com/recipes/578078/ (python LRU cache
# implementation).
def make_call_key(args, kwargs):
    """Stable id for function calls, can be used for caching."""
    key = args
    if kwargs:
        key += "##kwargs##", tuple(sorted(kwargs.items()))
    return key


@currying.curry
def top(iterable: Iterable, key=operator.identity):
    """Generates elements from max to min.

    >>> tuple(top((1, 3, 2)))
    (3, 2, 1)

    >>> tuple(top(('a', 'aa', 'aaa'), len))
    ('aaa', 'aa', 'a')
    """
    h: List = []
    for i, value in enumerate(iterable):
        # Use the index as a tie breaker.
        heapq_max.heappush_max(h, (key(value), i, value))
    while h:
        yield operator.nth(2)(heapq_max.heappop_max(h))


@currying.curry
def bottom(iterable, key=operator.identity):
    """Generates elements from min to max.

    >>> tuple(bottom((3, 2, 1)))
    (1, 2, 3)

    >>> tuple(bottom((1, 2, 3, 4), lambda x: x % 2 == 0))
    (1, 3, 2, 4)
    """
    h = []
    for i, value in enumerate(iterable):
        # Use the index as a tie breaker.
        heapq.heappush(h, (key(value), i, value))
    while h:
        yield operator.nth(2)(heapq.heappop(h))


def skip(n: int):
    """Skip the first n elements of a sequence. i.e, Return a generator that yields all elements after the n'th element.
    >>> tuple(skip(3)([i for i in range(6)]))
    (3, 4, 5)
    """

    def skip(seq: Iterable):
        for i, x in enumerate(seq):
            if i < n:
                continue
            yield x

    return skip


@currying.curry
def assoc_in(d, keys, value, factory=dict):
    """Associate a value to the input dict given the path "keys".

    >>> assoc_in({"a": {"b": 1}}, ["a", "b"], 2)
    {'a': {'b': 2}}
    """
    return update_in(d, keys, lambda x: value, value, factory)


@currying.curry
def update_in(d: dict, keys: Iterable, func: Callable, default=None, factory=dict):
    """Gets a (potentially nested) dictionary, key(s) and a function, and return new `dictionary` d' where d'[key] = func(d[key]).

    >>> inc = lambda x: x + 1
    >>> update_in({'a': 0}, ['a'], inc)
    {'a': 1}
    """
    ks = iter(keys)
    k = next(ks)

    rv = inner = factory()
    rv.update(d)

    for key in ks:
        if k in d or isinstance(d, list):
            d = d[k]
            if isinstance(d, dict):
                dtemp = {}
                dtemp.update(d)
            elif isinstance(d, list):
                dtemp = []
                dtemp.extend(d)
            else:
                dtemp = factory()
        else:
            d = dtemp = factory()

        inner[k] = inner = dtemp
        k = key

    if k in d:
        inner[k] = func(d[k])
    else:
        inner[k] = func(default)
    return rv


def dataclass_transform(
    attr_name: Text,
    attr_transformer: Callable[[Any], Any],
):
    """Return a new instance of the dataclass where new_dataclass_instance.attr_name = attr_transformer(dataclass_instance.attr_name)
    >>> @dataclasses.dataclass(frozen=True)
    ... class C:
    ...    x: int
    >>> c = C(5)
    >>> d = dataclass_transform('x', lambda i: i * 2)(c)
    >>> assert d.x == 10
    """
    transformation = sync.compose_left(
        operator.attrgetter(attr_name),
        attr_transformer,
    )

    def dataclass_transform(dataclass_instance):
        return dataclasses.replace(
            dataclass_instance,
            **{
                attr_name: transformation(dataclass_instance),
            },
        )

    return dataclass_transform


dataclass_transform_attribute = sync.binary_curry(dataclass_transform)


def dataclass_replace(attr_name: Text, attr_value) -> Callable:
    return dataclass_transform(attr_name, lambda _: attr_value)


dataclass_replace_attribute = sync.binary_curry(dataclass_replace)

_R = TypeVar("_R")
_E = TypeVar("_E")


@currying.curry
def reduce(
    reducer: Callable[[_R, _E], _R],
    initial_value: _R,
    elements: Iterable[_E],
) -> _R:
    return functools.reduce(reducer, elements, initial_value)


@currying.curry
def suffix(val: Any, it: Iterable):
    """Add a value to the end of an iterable. Return an iterable.

    >>> tuple(suffix(4, (1, 2, 3)))
    (1, 2, 3, 4)
    """
    return itertools.chain(it, (val,))


@currying.curry
def prefix(val: Any, it: Iterable):
    """Add a value to the beginning of an iterable. Return an iterable.

    >>> tuple(prefix(1, (2, 3, 4)))
    (1, 2, 3, 4)
    """
    return itertools.chain((val,), it)


@currying.curry
def concat_with(new_it: Iterable, it: Iterable):
    """Concat two iterables.

    >>> tuple(concat_with((3, 4), (1, 2)))
    (1, 2, 3, 4)
    """
    return itertools.chain(it, new_it)


@currying.curry
def drop_last_while(predicate: Callable[[Any], bool], seq: Sequence) -> Sequence:
    return sync.pipe(
        seq,
        reversed,
        currying.curry(itertools.dropwhile)(predicate),
        tuple,
        reversed,
    )


def take_while(predicate: Callable[[Any], bool]):
    """Take elements from an iterable as long as elements pass some predicate.

    >>> list(functional.take_while(lambda x: x < 7)([1, 2, 9, 2]))
    [1, 2]
    """

    def take_while(iterable: Iterable) -> Iterable:
        for x in iterable:
            if not predicate(x):
                break
            yield x

    return take_while


@currying.curry
def partition_after(
    predicate: Callable[[Any], bool],
    seq: Sequence,
) -> Sequence[Sequence]:
    return toolz.reduce(
        lambda a, b: (*a, (b,))
        if not a or predicate(a[-1][-1])
        else (*a[:-1], (*a[-1], b)),
        seq,
        (),
    )


@currying.curry
def partition_before(
    predicate: Callable[[Any], bool],
    seq: Sequence,
) -> Sequence[Sequence]:
    return toolz.reduce(
        lambda a, b: (*a, (b,)) if not a or predicate(b) else (*a[:-1], (*a[-1], b)),
        seq,
        (),
    )


def get_all_n_grams(seq: Sequence) -> Iterable[Tuple]:
    for i in range(len(seq)):
        for j in range(i + 1, len(seq) + 1):
            yield tuple(seq[i:j])


@currying.curry
def eq_by(f, value_1, value_2):
    """Check if two values are equal when applying f on both of them."""
    return f(value_1) == f(value_2)


#: Check if two strings are equal, ignoring case.
#: >>> eq_str_ignore_case("HeLlO wOrLd", "hello world")
#: True
eq_str_ignore_case = eq_by(str.lower)


@currying.curry
def groupby_many_reduce(key: Callable, reducer: Callable, seq: Iterable):
    """Group a collection by a key function, when the value is given by a reducer function.

    Parameters:
    key (Callable): Key function (given object in collection outputs key).
    reducer (Callable): Reducer function (given object in collection outputs new value).
    seq (Iterable): Collection.

    Returns:
    Dict[Text, Any]: Dictionary where key has been computed by the `key` function
    and value by the `reducer` function.

    >>> groupby_many_reduce(head, lambda x, y: x + len(y) if x else len(y), ["hello", "hi", "test", "to"])
    {'h': 7, 't': 6}
    """
    result: Dict[Any, Any] = {}
    for element in seq:
        for key_result in key(element):
            result[key_result] = reducer(result.get(key_result, None), element)
    return result


def unique_by(f):
    """Return only unique elements of a sequence defined by function f

    >>> tuple(unique_by(len)(['cat', 'mouse', 'dog', 'hen']))
    ('cat', 'mouse')
    """

    def unique(seq):
        seen = set()
        for item in seq:
            val = f(item)
            if val not in seen:
                seen.add(val)
                yield item

    return unique


#: Return only unique elements of a sequence
#:
#: >>> tuple(unique(["cat", "mouse", "dog", "cat"]))
#: ('cat', 'mouse', 'dog')
unique = unique_by(operator.identity)


def interpose(el):
    """Introduces an element between each pair of elements in the input sequence.

    >>> tuple(interpose("a")([1, 2, 3]))
    (1, 'a', 2, 'a', 3)
    """

    def interpose_inner(seq):
        return toolz.interpose(el, seq)

    return interpose_inner


def take(n: int):
    """Get an iterator for the first n elements of a sequence.

    >>> tuple(take(3) ([1, 2, 3, 4, 5]))
    (1, 2, 3)
    """

    def take(seq):
        return itertools.islice(seq, n)

    return take


def drop(n: int):
    """Drops the first n elements of a sequence.

    >>> tuple(drop(2)([1,2,3,4,5]))
    (3,4,5)
    """

    def drop(seq):
        return itertools.islice(seq, n, None)

    return drop


def drop_last(n: int):
    """Drops the last n elements of a sequence.

    >>> tuple(drop_last(1)([1,2,3,4,5]))
    (1,2,3,4)
    """

    def drop_last(seq):
        return itertools.islice(seq, len(seq) - n)

    return drop_last


def flip(func: Callable):
    """Call the function call with the arguments flipped.

    >>> import operator; flip(operator.truediv)(2, 6)
    3.0
    """

    @currying.curry
    def flip(a, b):
        return func(b, a)

    return flip


def sliding_window(n: int):
    """A sequence of overlapping subsequences.

    >>> list(sliding_window(2)([1, 2, 3, 4]))
    [(1, 2), (2, 3), (3, 4)]
    """

    def sliding_window(seq):
        return toolz.sliding_window(n, seq)

    return sliding_window


def partition_all(n: int):
    """Partition all elements of sequence into tuples of length at most n.
    The final tuple may be shorter to accommodate extra elements.

    >>> list(partition_all(2)([1, 2, 3, 4]))
    [(1, 2), (3, 4)]

    >>> list(partition_all(2)([1, 2, 3, 4, 5]))
    [(1, 2), (3, 4), (5,)]
    """

    def partition_all(seq):
        return toolz.partition_all(n, seq)

    return partition_all


def ends_with(expected_tail: Iterable) -> Callable[[Sequence], bool]:
    """Returns a predicate that checks if an iterabel ends with another iterable.

    >>> ends_with([1,2,3])((0,1,2,3))
    True
    >>> ends_with([1,2,3])((1,2))
    False
    >>> ends_with([1,2])((3,1,2))
    True
    >>> ends_with([1])(())
    False
    """

    class Nothing:
        pass

    expected_tail_as_tuple = tuple(expected_tail)

    def ends_with(seq: Iterable):
        tail_of_seq = operator.tail(len(expected_tail_as_tuple))(seq)
        for a, b in itertools.zip_longest(
            tail_of_seq,
            expected_tail_as_tuple,
            fillvalue=Nothing(),
        ):
            if a != b:
                return False
        return True

    return ends_with


def intersect(collections: Collection[Collection]) -> Iterable:
    """Intersects a group of collections.

    Caller is responsible to give collections with O(1) containment checks.

    >>> tuple(
            gamla.intersect(
                [
                    [1, 2, 3, 4],
                    [4, 5],
                    [4, 2],
                ]
            )
        )
    (4,)

    """
    first_collection, *rest = collections
    for x in first_collection:
        if all(x in container for container in rest):
            yield x


have_intersection = sync.compose_left(intersect, operator.nonempty)


def function_to_uid(f: Callable) -> str:
    """Returns a unique identifier for the given function."""
    return hashlib.sha1(f.__name__.encode("utf-8")).hexdigest()


#: Directory path of a given function
function_to_directory = sync.compose_left(
    operator.attrgetter("__code__"),
    operator.attrgetter("co_filename"),
    os.path.dirname,
)


def function_and_input_to_identifier(factory) -> Callable:
    """Returns a unique identifier for the given function and input."""

    def inner(args, kwargs) -> str:
        return sync.pipe(
            (
                function_to_uid(factory),
                compute_stable_json_hash(make_call_key(args, kwargs)),
            ),
            sync.filter(operator.identity),
            "-".join,
        )

    return inner


#: Average of an iterable. If the sequence is empty, returns 0.
#: >>> average([1,2,3])
#: 2.0
average = sync.compose_left(
    sync.bifurcate(sum, operator.count),
    excepts_decorator.excepts(
        ZeroDivisionError,
        construct.just(0),
        sync.star(truediv),
    ),
)


def attr_equals(attribute: str, equals_what: Any) -> Callable[[Any], bool]:
    """Returns a function that get an object x and returns whether x.attribute == equals_what

    >>> attr_equals("imag", 5.0)(8 + 5j)
    True
    >>> attr_equals("imag", 5.0)(8)
    False
    """
    return sync.compose_left(
        operator.attrgetter(attribute),
        operator.equals(equals_what),
    )


def sample_with_randint(randint: Callable, k: int):
    """Samples an iterable uniformly in one pass with O(k) memory.

    >>> sample(2)([1, 2, 3])
    frozenset([1,3])
    """

    def reducer(
        index_and_sample: Tuple[int, immutables.Map],
        current,
    ) -> Tuple[int, immutables.Map]:
        index, sample = index_and_sample
        replacement_index = index if index < k else randint(0, index)
        return (
            index + 1,
            sample.set(replacement_index, current) if replacement_index < k else sample,
        )

    return sync.compose_left(
        reduce(reducer, (0, immutables.Map())),
        operator.second,
        immutables.Map.values,
        frozenset,
    )


sample = currying.curry(sample_with_randint)(random.randint)
choice = sync.compose(operator.head, sample_with_randint(random.randint, 1))



import dataclasses
import itertools
import json
from typing import Callable, Collection, Dict

import dataclasses_json

from gamla import construct, functional_generic, operator
from gamla.optimized import sync


def _immutable(self, *args, **kws):
    raise TypeError("cannot change object - object is immutable")


class frozendict(dict):  # noqa: N801
    def __init__(self, *args, **kwargs):
        self.__setattr__ = _immutable
        super(frozendict, self).__init__(*args, **kwargs)

    def __hash__(self):
        return hash(tuple(self.items()))

    def __gt__(self, other):
        return functional_generic.map_dict(dict.items, operator.identity)(
            self,
        ) > functional_generic.map_dict(dict.items, operator.identity)(other)

    # TODO(nitzo): Disabled since we need to be able to un-serialize with dill/pickle.
    # __setitem__ = _immutable
    __delitem__ = _immutable
    pop = _immutable
    popitem = _immutable
    clear = _immutable
    update = _immutable
    setdefault = _immutable


def get_encode_config():
    """Display dataclass field as a tuple of JSON strings."""
    return dataclasses.field(
        metadata=dataclasses_json.config(
            encoder=lambda lst: sorted(lst, key=json.dumps, reverse=False),
        ),
    )


def _freeze_nonterminal(v):
    if isinstance(v, Dict):
        return frozendict(v)
    return tuple(v)


#: Freeze recursively a dictionary.
#:
#: >>> freeze_deep({"1": {"2": "345", "some-string": ["hello"]}})
#: data.frozendict(
#:  {"1": data.frozendict({"2": "345", "some-string": ("hello",)})},
#: )
freeze_deep = functional_generic.map_dict(_freeze_nonterminal, operator.identity)


class Enum(frozenset):
    def __getattr__(self, name):
        if name in self:
            return name
        raise AttributeError


def _do_on_positions(f, predicate: Callable[[int], bool]):
    return sync.compose_left(
        enumerate,
        sync.map(
            sync.ternary(
                sync.compose_left(
                    operator.head,
                    predicate,
                ),
                sync.compose_left(operator.second, f),
                operator.second,
            ),
        ),
    )


def explode(*positions: Collection[int]):
    """Flattens a non homogeneous iterable.

    For an iterable where some positions are iterable and some are not,
    "explodes" the iterable, so that each element appears in a single row, and duplicates the non iterable.

    >>> functional_generic.pipe(
    ...     ["x", ["y1", "y2", "y3"], "z"],
    ...     data.explode(1),
    ...     tuple,
    ... )
    (
        ("x", "y1", "z"),
        ("x", "y2", "z"),
        ("x", "y3", "z"),
    )
    """
    return sync.compose_left(
        _do_on_positions(
            construct.wrap_tuple,
            sync.complement(operator.contains(positions)),
        ),
        sync.star(itertools.product),
    )


def transform_if_not_none(transform: Callable, value):
    """
    Apply a function on a given value if it's not None. Else, return the None value.

    >>> transform_if_not_none(
    ...     functional_generic.when(operator.is_instance, lambda x: x.casefold())),
    ...     "Some Title"
    ... )
    'some title'
    >>> transform_if_not_none(
    ...     functional_generic.when(operator.is_instance, lambda x: x.casefold())),
    ...     None
    ... )

    """
    if value is not None:
        return transform(value)
    return value



from typing import Text


def apply(*args, **kwargs):
    """Apply input on function.

    >>> apply(1)(add(2))
    3
    """

    def apply_inner(function):
        return function(*args, **kwargs)

    return apply_inner


def apply_async(*args, **kwargs):
    """Apply input on an async function.

    >>> apply_async(1)(async_add(2))
    3
    """

    async def apply_async(f):
        return await f(*args, **kwargs)

    return apply_async


def apply_fn_with_args(fn, *args):
    """Returns the result of applying `fn(*args)`."""
    return fn(*args)


def apply_method_async(method: Text, *args, **kwargs):
    """Invokes the specified async method on an object with `*args` and `**kwargs`.

    >>> apply_method_async("get", "http://www.someurl.com")(httpx)
    httpx.Response()
    """

    async def apply_method_async(obj):
        return await apply_async(*args, **kwargs)(getattr(obj, method))

    return apply_method_async


def apply_method(method: Text, *args, **kwargs):
    """Invokes the specified method on an object with `*args` and `**kwargs`.

    >>> apply_method("get", "http://www.someurl.com")(requests)
    requests.Response()
    """

    def apply_method(obj):
        return apply(*args, **kwargs)(getattr(obj, method))

    return apply_method


def invoke(f):
    """Performs a call of the input function.
    >>> invoke(lambda: 0)
    0
    """
    return f()



import asyncio
import builtins
import functools
import logging
from typing import Text

import tabulate
import termcolor
import toolz
import yappi

from gamla import functional_generic
from gamla.optimized import sync

logger = functional_generic.side_effect(logging.info)


def log_text(text: Text, level: int = logging.INFO):
    """Logs given text and returns the execution input,
    can also log the input itself by using {} in the text string.

    >>> f = log_text("hello world {}")
    >>> f("It's me!")
    INFO hello world It's me!
    'It's me!'
    """
    return functional_generic.side_effect(lambda x: logging.log(level, text.format(x)))


def _log_and_break(x):
    logging.info(termcolor.colored(x, "yellow"))
    builtins.breakpoint()


#: A util to inspect a pipline by opening a debug prompt.
#: Note:
#: - Materializes generators, as most of the time we are interested in looking into them, so can have unexpected results.
#: - The current value can be referenced by `x` in the debug prompt.
debug = functional_generic.compose_left(
    functional_generic.when(functional_generic.is_generator, tuple),
    functional_generic.side_effect(_log_and_break),
)

debug_after = functional_generic.after(debug)
debug_before = functional_generic.before(debug)


def _debug_generic(f):
    return functional_generic.compose_left(
        lambda *funcs: toolz.interleave([funcs, [debug] * len(funcs)]),
        sync.star(f),
    )


#: Replace regular `compose` calls with this to get a breakpoint at each step.
debug_compose = _debug_generic(functional_generic.compose)
#: Replace regular `compose_left` calls with this to get a breakpoint at each step.
debug_compose_left = _debug_generic(functional_generic.compose_left)


def debug_exception(f):
    """Debug exception in a pipeline stage by looking at the causal value.

    >>> gamla.pipe(
    ...     "abc",
    ...     gamla.itemgetter("some_key"),  # This would cause an exception.
    ... )

    >>> gamla.pipe(
    ...     "abc",
    ...     gamla.debug_exception(gamla.itemgetter("some_key")),  # Now we can see the cause of the exception - we expect a `dict` but get a `str`.
    ... )
    """
    if asyncio.iscoroutinefunction(f):

        async def debug_exception(*x, **kwargs):
            try:
                return await f(*x, **kwargs)
            except Exception as e:
                builtins.breakpoint()
                raise e

    else:

        def debug_exception(*x, **kwargs):  # type: ignore
            try:
                return f(*x, **kwargs)
            except Exception as e:
                builtins.breakpoint()
                raise e

    return functools.wraps(f)(debug_exception)


def _print_stats():
    logging.info(
        "\n\n\n"
        + tabulate.tabulate(
            tuple(
                map(
                    lambda stat: [
                        f"{stat.ttot:2f}",
                        f"{stat.tsub:2f}",
                        stat.ncall,
                        stat.name,
                        f"{stat.module}:{stat.lineno}",
                    ],
                    yappi.get_func_stats(),
                ),
            )[:100],
            [
                "ttot",
                "tsub",
                "ncall",
                "name",
                "module:lineno",
            ],
        ),
    )


def profileit(f):
    if asyncio.iscoroutinefunction(f):

        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            yappi.set_clock_type("WALL")
            with yappi.run():
                result = await f(*args, **kwargs)
            _print_stats()
            return result

        return wrapper

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        yappi.set_clock_type("WALL")
        with yappi.run():
            result = f(*args, **kwargs)
        _print_stats()
        return result

    return wrapper



import asyncio
import random
import time
from operator import not_, truediv

import pytest

from gamla import (
    construct,
    currying,
    dict_utils,
    functional,
    functional_generic,
    operator,
)
from gamla.optimized import sync


async def _opposite_async(x):
    await asyncio.sleep(0.01)
    return not x


async def _equals(x, y):
    await asyncio.sleep(0.01)
    return x == y


def test_do_if():
    assert functional.do_if(construct.just(True), construct.just(2))(1) == 1


def test_currying():
    @currying.curry
    def f(x, y, z):
        return x + y + z

    assert f(1, 2, 3) == 6
    assert f(1)(2, 3) == 6
    assert f(1, 2)(3) == 6


def test_juxt():
    assert functional_generic.juxt(operator.identity, operator.add(1))(3) == (3, 4)


def test_juxt_zero_params():
    assert functional_generic.juxt(lambda: 1, lambda: 3)() == (1, 3)


async def test_juxt_zero_params_async():
    async def slow_1():
        await asyncio.sleep(0.01)
        return 1

    assert await functional_generic.juxt(lambda: 3, slow_1)() == (3, 1)


async def test_juxt_async():
    async def slow_identity(x):
        await asyncio.sleep(0.01)
        return x

    assert await functional_generic.juxt(operator.identity, slow_identity)(3) == (
        3,
        3,
    )


def test_anyjuxt():
    assert functional_generic.anyjuxt(not_, operator.identity)(True)


def test_alljuxt():
    assert not functional_generic.alljuxt(not_, operator.identity)(True)


async def test_alljuxt_async():
    assert not await functional_generic.alljuxt(_opposite_async, operator.identity)(
        True,
    )


async def test_anyjuxt_async():
    assert await functional_generic.anyjuxt(_opposite_async, operator.identity)(True)


async def test_anymap():
    assert await functional_generic.anymap(_opposite_async)([True, True, False])


async def test_allmap():
    def opposite(x):
        return not x

    assert not functional_generic.allmap(opposite)([True, True, False])


async def test_anymap_async():
    assert await functional_generic.anymap(_opposite_async)([True, True, False])


async def test_allmap_async():
    assert not await functional_generic.allmap(_opposite_async)([True, True, False])


async def test_allmap_in_async_pipe():
    assert not await functional_generic.pipe(
        [True, True, False],
        functional_generic.allmap(_opposite_async),
        # Check that the `pipe` serves a value and not a future.
        sync.check(operator.is_instance(bool), AssertionError),
    )


async def test_anymap_in_pipe():
    assert not functional_generic.pipe(
        [True, True, False],
        functional_generic.allmap(not_),
    )


async def test_itemmap_async_sync_mixed():
    assert await functional_generic.pipe(
        {True: True},
        functional_generic.itemmap(
            functional_generic.compose(
                tuple,
                functional_generic.curried_map(_opposite_async),
            ),
        ),
        functional_generic.itemmap(
            functional_generic.compose(
                tuple,
                functional_generic.curried_map(not_),
            ),
        ),
    ) == {True: True}


async def test_itemfilter_async_sync_mixed():
    assert await functional_generic.pipe(
        {1: 1, 2: 1, 3: 3},
        functional_generic.itemfilter(functional_generic.star(_equals)),
        functional_generic.itemfilter(sync.star(lambda _, val: val == 1)),
    ) == {1: 1}


async def test_keymap_async_curried():
    assert await functional_generic.keymap(_opposite_async)({True: True}) == {
        False: True,
    }


async def test_keyfilter_sync_curried():
    assert functional_generic.keyfilter(operator.identity)(
        {False: True, True: False},
    ) == {True: False}


async def test_valmap_sync_curried():
    assert functional_generic.valmap(not_)({True: True}) == {True: False}


async def test_valfilter_sync_curried():
    assert functional_generic.valfilter(operator.identity)(
        {False: True, True: False},
    ) == {False: True}


async def _is_even_async(x):
    await asyncio.sleep(0.1)
    return x % 2 == 0


async def test_filter_curried_async_sync_mix():
    assert await functional_generic.pipe(
        [1, 2, 3, 4],
        functional_generic.curried_filter(_is_even_async),
        functional_generic.curried_map(operator.add(10)),
        tuple,
    ) == (12, 14)


def test_case_single_predicate():
    assert functional_generic.case_dict({operator.identity: operator.identity})(
        True,
    )


def test_case_multiple_predicates():
    assert not functional_generic.case_dict(
        {not_: operator.identity, operator.identity: not_},
    )(True)


def test_case_no_predicate():
    with pytest.raises(sync.NoConditionMatched):
        functional_generic.case_dict(
            {
                not_: operator.identity,
                # Can't repeat keys.
                lambda x: not x: operator.identity,
            },
        )(True)


async def test_case_async():
    assert not await functional_generic.case_dict(
        {_opposite_async: operator.identity, operator.identity: _opposite_async},
    )(True)


def test_partition_after():
    assert functional.partition_after(operator.equals(1), []) == ()
    assert tuple(
        functional.partition_after(
            operator.equals(1),
            [1, 1, 2, 2, 1, 1, 2, 1, 1, 1],
        ),
    ) == ((1,), (1,), (2, 2, 1), (1,), (2, 1), (1,), (1,))


def test_partition_before():
    assert functional.partition_before(operator.equals(1), []) == ()
    assert tuple(
        functional.partition_before(
            operator.equals(1),
            [3, 1, 1, 2, 2, 1, 1, 2, 1, 1, 1],
        ),
    ) == ((3,), (1,), (1, 2, 2), (1,), (1, 2), (1,), (1,), (1,))


async def test_drop_last_while():
    assert tuple(functional.drop_last_while(operator.equals(1), [])) == ()
    assert tuple(functional.drop_last_while(operator.equals(1), [1])) == ()
    assert tuple(functional.drop_last_while(operator.equals(1), [2])) == (2,)
    assert tuple(
        functional.drop_last_while(
            operator.equals(1),
            [1, 1, 2, 2, 1, 1, 2, 1, 1, 1],
        ),
    ) == (1, 1, 2, 2, 1, 1, 2)


def test_apply_spec():
    assert functional_generic.apply_spec(
        {"identity": operator.identity, "increment": operator.add(1)},
    )(1) == {"identity": 1, "increment": 2}


async def test_apply_spec_async():
    async def async_identity(x):
        await asyncio.sleep(0.01)
        return x

    assert await functional_generic.apply_spec(
        {"identity": async_identity, "increment": operator.add(1)},
    )(1) == {"identity": 1, "increment": 2}


async def test_apply_spec_async_recursive():
    async def async_identity(x):
        await asyncio.sleep(0.01)
        return x

    f = functional_generic.apply_spec(
        {"identity": {"nested": async_identity}, "increment": operator.add(1)},
    )
    assert await f(1) == {"identity": {"nested": 1}, "increment": 2}


async def test_async_bifurcate():
    async def async_sum(x):
        await asyncio.sleep(0.01)
        return sum(x)

    def gen():
        yield 1
        yield 2
        yield 3

    average = await functional_generic.pipe(
        gen(),
        functional_generic.bifurcate(async_sum, operator.count),
        sync.star(truediv),
    )

    assert average == 2


def test_average():
    def gen():
        yield 1
        yield 2
        yield 3

    assert functional.average(gen()) == 2


def test_countby_many():
    names = ["alice", "bob", "charlie", "dan", "edith", "frank"]
    assert functional_generic.countby_many(lambda name: (name[0], name[-1]))(names) == {
        "a": 1,
        "e": 3,
        "b": 2,
        "c": 1,
        "d": 1,
        "n": 1,
        "h": 1,
        "f": 1,
        "k": 1,
    }


async def test_reduce_async():
    async def slow_addition(x, y):
        await asyncio.sleep(0.1)
        return x + y

    assert await (functional_generic.reduce_curried(slow_addition, 0)([1, 2, 3])) == 6


def test_reduce():
    def addition(x, y):
        return x + y

    assert functional_generic.reduce_curried(addition, 0)([1, 2, 3]) == 6


async def test_scan():
    def addition(x, y):
        return x + y

    assert (functional_generic.scan(addition, 0)([1, 2, 3])) == (
        0,
        1,
        3,
        6,
    )


async def test_scan_async():
    async def slow_addition(x, y):
        await asyncio.sleep(0.01)
        return x + y

    assert await (functional_generic.scan(slow_addition, 0)([1, 2, 3])) == (
        0,
        1,
        3,
        6,
    )


def test_find():
    seq = ({"key": 1}, {"key": 2}, {"key": 3}, {"key": 2})

    assert functional_generic.find(
        functional_generic.compose_left(
            dict_utils.itemgetter("key"),
            operator.equals(2),
        ),
    )(iter(seq)) == {"key": 2}

    assert (
        functional_generic.find(
            functional_generic.compose_left(
                dict_utils.itemgetter("key"),
                operator.equals(4),
            ),
        )(iter(seq))
        is None
    )


def test_find_index():
    seq = ({"key": 1}, {"key": 2}, {"key": 3}, {"key": 2})

    assert (
        functional_generic.find_index(
            functional_generic.compose_left(
                dict_utils.itemgetter("key"),
                operator.equals(2),
            ),
        )(iter(seq))
        == 1
    )

    assert (
        functional_generic.find_index(
            functional_generic.compose_left(
                dict_utils.itemgetter("key"),
                operator.equals(4),
            ),
        )(iter(seq))
        == -1
    )


def test_compositions_have_name():
    assert (
        functional_generic.compose_left(
            operator.identity,
            operator.identity,
            functional.unique,
        ).__name__
        == "unique_OF_identity_OF_identity"
    )


def test_async_compositions_have_name():
    async def async_identity(x):
        await asyncio.sleep(0)
        return x

    assert (
        functional_generic.compose_left(
            operator.identity,
            async_identity,
            functional.unique,
        ).__name__
        == "unique_OF_async_identity_OF_identity"
    )


def test_compositions_correct_typing():
    def f(x: str) -> str:
        return "hello " + x

    def g(x: str) -> int:
        return len(x)

    assert functional_generic.compose_left(f, g).__annotations__ == {
        "x": str,
        "return": int,
    }


def test_typing_for_composition_with_builtin():
    def f(x: str) -> str:
        return "hello " + x

    assert functional_generic.compose_left(f, bool).__annotations__ == {
        "x": str,
    }


def test_attrgetter():
    assert operator.attrgetter("lower")("ASD")() == "asd"


def test_attr_equals():
    assert functional.attr_equals("imag", 5.0)(8 + 5j) is True
    assert functional.attr_equals("real", 5.0)(8 + 5j) is False


def test_latency():
    start_time = time.time()
    for _ in range(1000):
        functional_generic.pipe(
            True,
            functional_generic.juxt(operator.equals(True), operator.equals(False)),
            operator.head,
            construct.just("bla"),
            operator.attrgetter("lower"),
        )
    assert time.time() - start_time < 0.1


def test_unique_by():
    assert tuple(
        functional.unique_by(lambda x: x[0])(["a", "ab", "abc", "bc", "c"]),
    ) == ("a", "bc", "c")
    assert tuple(functional.unique(["a", "a", "a", "bc", "a"])) == ("a", "bc")


def test_merge():
    assert functional_generic.merge(
        {"1": 1, "2": 2},
        {"2": 3, "3": 3},
    ) == {"1": 1, "2": 3, "3": 3}


def test_merge_with():
    assert functional_generic.merge_with(operator.head)(
        {"1": 1, "2": 2},
        {"2": 3, "3": 3},
    ) == {"1": 1, "2": 2, "3": 3}


async def test_async_merge_with():
    async def async_first(x):
        await asyncio.sleep(0.01)
        return x[0]

    assert await functional_generic.merge_with(async_first)(
        {"1": 1, "2": 2},
        {"2": 3, "3": 3},
    ) == {"1": 1, "2": 2, "3": 3}


async def test_async_when():
    async def async_equals_1(x):
        await asyncio.sleep(0.01)
        return x == 1

    assert await functional_generic.when(async_equals_1, construct.just(True))(1)

    assert await functional_generic.when(async_equals_1, construct.just(True))(2) == 2


async def test_unless1():
    assert functional_generic.unless(operator.equals(1), construct.just(True))(1) == 1


async def test_unless2():
    assert functional_generic.unless(operator.equals(1), construct.just(True))(2)


def test_compose_many_to_one():
    assert (
        functional_generic.compose_many_to_one([sum, sum], lambda x, y: x + y)(
            [1, 2, 3],
        )
        == 12
    )


def test_empty_pipe():
    with pytest.raises(functional_generic.PipeNotGivenAnyFunctions):
        functional_generic.pipe(
            [
                1,
                2,
                3,
            ],
        )


def test_groupby():
    names = ["alice", "bob", "charlie", "dan", "edith", "frank"]
    assert functional_generic.groupby(operator.last)(names) == {
        "e": ("alice", "charlie"),
        "b": ("bob",),
        "n": ("dan",),
        "h": ("edith",),
        "k": ("frank",),
    }


def test_groupby_empty():
    assert functional_generic.groupby(operator.last)([]) == {}


def test_take():
    assert tuple(functional.take(2)(["a", "b", "c"])) == ("a", "b")


def test_drop():
    assert tuple(functional.drop(2)(["a", "b", "c"])) == ("c",)


def test_drop_last():
    assert tuple(functional.drop_last(1)(["a", "b", "c"])) == ("a", "b")


def test_count_by():
    assert functional_generic.count_by(operator.head)(["aa", "ab", "ac", "bc"]) == {
        "a": 3,
        "b": 1,
    }


def test_add():
    assert operator.add(1)(2) == 3
    assert operator.add(["c"])(["a", "b"]) == ["a", "b", "c"]


def test_assert_that():
    functional.assert_that(operator.equals(2))(2)


def test_assert_that_with_message():
    functional.assert_that_with_message(
        construct.just("Input is not 2!"),
        operator.equals(2),
    )(2)
    with pytest.raises(AssertionError, match="Input is not 2!"):
        functional.assert_that_with_message(
            construct.just("Input is not 2!"),
            operator.equals(2),
        )(3)


def test_assoc_in():
    assert functional.assoc_in({"a": {"b": 1}}, ["a", "b"], 2) == {"a": {"b": 2}}


def test_bottom():
    assert tuple(functional.bottom((3, 2, 1))) == (1, 2, 3)
    assert tuple(functional.bottom((1, 2, 3, 4), lambda x: x % 2 == 0)) == (1, 3, 2, 4)


def test_concat_with():
    assert tuple(functional.concat_with((3, 4), (1, 2))) == (1, 2, 3, 4)


def test_contains():
    assert operator.contains([1, 2, 3])(2)
    assert not operator.contains("David")("x")


def test_side_effect():
    side_result = []
    side_effect = functional_generic.compose_left(
        operator.multiply(2),
        side_result.append,
    )
    assert functional_generic.side_effect(side_effect)(2) == 2
    assert side_result == [4]


async def test_side_effect_async():
    side_result = []
    side_effect = functional_generic.compose_left(
        currying.curry(_equals)(2),
        side_result.append,
    )
    assert await functional_generic.side_effect(side_effect)(2) == 2
    assert side_result == [True]


def test_side_effect_not_exhausting_generator():
    side_result = []
    gen = functional_generic.pipe(
        [1, 2, 3],
        functional_generic.curried_map(lambda x: x + 1),
        functional_generic.side_effect(
            functional_generic.compose_left(tuple, len, side_result.append),
        ),
    )
    assert side_result == [3]
    assert tuple(gen)


def test_sliding_window():
    assert list(functional.sliding_window(2)([1, 2, 3, 4])) == [(1, 2), (2, 3), (3, 4)]


def test_partition_all():
    assert list(functional.partition_all(2)([1, 2, 3, 4])) == [(1, 2), (3, 4)]
    assert list(functional.partition_all(2)([1, 2, 3, 4, 5])) == [(1, 2), (3, 4), (5,)]


def test_ends_with():
    assert functional.ends_with([1, 2, 3])((0, 1, 2, 3)) is True
    assert functional.ends_with([1, 2, 3])((1, 2)) is False
    assert functional.ends_with([1, 2])((3, 1, 2)) is True
    assert functional.ends_with([1])(()) is False
    assert functional.ends_with([1, 2, 3])((0, 1, 2, 3))
    assert functional.ends_with(range(1, 4))(range(0, 4))
    assert not functional.ends_with([1, 2, 3])((1, 2))
    assert functional.ends_with([1, 2])((3, 1, 2))
    assert not functional.ends_with([1])(())


def test_flip():
    assert functional.flip(truediv)(2, 6) == 3.0


def test_intersect():
    assert tuple(functional.intersect([[1, 2, 3, 4], [4, 5], [4, 2]])) == (4,)


def test_has_intersection():
    assert functional.have_intersection([(1, 2, 3), (3, 4, 5)]) is True


def test_doesnt_have_intersection():
    assert functional.have_intersection([(1, 2, 3), (4, 5, 6)]) is False


def test_all_n_grams():
    assert set(functional.get_all_n_grams("abc")) == {
        ("a",),
        ("a", "b"),
        (
            "a",
            "b",
            "c",
        ),
        ("b",),
        ("b", "c"),
        ("c",),
    }


def test_all_n_grams_non_textual():
    assert set(functional.get_all_n_grams([1, 2, 3])) == {
        (1,),
        (1, 2),
        (
            1,
            2,
            3,
        ),
        (2,),
        (2, 3),
        (3,),
    }


def test_take_while():
    assert list(functional.take_while(lambda x: x < 7)([1, 2, 9, 2])) == [1, 2]


def test_translate_exception():
    with pytest.raises(ValueError):
        functional_generic.pipe(
            iter([]),
            functional.translate_exception(next, StopIteration, ValueError),
        )


def test_speed_of_apply_spec():
    start_time = time.time()
    for _ in range(1000):
        functional_generic.apply_spec({"a": operator.identity, "b": lambda x: x + 1})(1)
    assert time.time() - start_time < 0.1


def test_compose_empty():
    assert functional_generic.compose()(1) == 1


def test_compose_single_function():
    assert functional_generic.compose(lambda x: x + 1)(1) == 2


def test_sample_with_randint():
    assert functional.sample_with_randint(random.Random(0).randint, 1)(
        range(10),
    ) == frozenset([3])



import asyncio
import functools
from typing import Any, Callable, Tuple, Union

from gamla import currying


@currying.curry
def excepts(
    exception: Union[Tuple[Exception, ...], Exception],
    handler: Callable[[Exception], Any],
    function: Callable,
):
    if asyncio.iscoroutinefunction(function):

        async def excepts(*args, **kwargs):
            try:
                return await function(*args, **kwargs)
            except exception as error:
                return handler(error)

    else:

        def excepts(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except exception as error:
                return handler(error)

    return functools.wraps(function)(excepts)


def try_and_excepts(
    exception: Union[Tuple[Exception, ...], Exception],
    handler: Callable,
    function: Callable,
):
    """Same as sync excepts only that the handler gets the original function params after the exception param."""
    if asyncio.iscoroutinefunction(function):

        async def try_and_excepts(*args, **kwargs):
            try:
                return await function(*args, **kwargs)
            except exception as error:
                return await handler(error, *args, **kwargs)

    else:

        def try_and_excepts(*args, **kwargs):
            try:
                return function(*args, **kwargs)
            except exception as error:
                return handler(error, *args, **kwargs)

    return functools.wraps(function)(try_and_excepts)



from gamla import operator


def _generator():
    yield 1
    yield 2
    yield 5


def test_nth():
    assert operator.nth(1)(["a", "b", "c"]) == "b"


def test_len_equals():
    assert operator.len_equals(3)(_generator())


def test_not_empty():
    assert operator.empty(_generator()) is False


def test_is_empty():
    assert operator.empty([]) is True


def test_nonempty():
    assert operator.nonempty(_generator()) is True


def test_between():
    assert operator.between(1, 100)(30) is True


def test_not_between():
    assert operator.between(2, 5)(14) is False



import asyncio
import dataclasses

from gamla import construct, excepts_decorator, functional, functional_generic, operator


@dataclasses.dataclass(frozen=True)
class _SomeError(Exception):
    pass


def test_excepts_sync():
    assert (
        excepts_decorator.excepts(
            _SomeError,
            construct.just(None),
            operator.identity,
        )(1)
        == 1
    )
    assert (
        excepts_decorator.excepts(
            _SomeError,
            construct.just(None),
            functional.make_raise(_SomeError),
        )(1)
        is None
    )


async def test_excepts_async():
    async def async_raise(x):
        raise _SomeError

    async def slow_identity(x):
        await asyncio.sleep(0.01)
        return x

    assert (
        await excepts_decorator.excepts(
            _SomeError,
            construct.just(None),
            slow_identity,
        )(1)
        == 1
    )
    assert (
        await excepts_decorator.excepts(
            _SomeError,
            construct.just(None),
            async_raise,
        )(1)
        is None
    )


def test_try_and_excepts_no_exception():
    assert (
        excepts_decorator.try_and_excepts(
            _SomeError,
            construct.just(None),
            operator.identity,
        )(1)
        == 1
    )


def test_try_and_excepts_with_exception():
    assert excepts_decorator.try_and_excepts(
        _SomeError,
        functional_generic.compose_left(operator.pack, operator.identity),
        functional.make_raise(_SomeError),
    )(1) == (_SomeError(), 1)



import asyncio
import time

from gamla import operator, tree


def test_get_leaves_by_ancestor_predicate():
    fn = tree.get_leaves_by_ancestor_predicate(operator.equals("x"))
    assert fn({"x": {"t": ("1")}}) == ("1",)


def test_get_leaves_by_ancestor_predicate_integer():
    fn = tree.get_leaves_by_ancestor_predicate(operator.less_than(4))
    assert fn({7: {2: ("bla")}}) == ("bla",)


def test_get_leaves_by_ancestor_predicate_no_matches():
    fn = tree.get_leaves_by_ancestor_predicate(operator.equals("x"))
    assert fn({"t": {"t": ("1")}}) == ()


def test_filter_leaves():
    fn = tree.filter_leaves(operator.greater_than(3))
    assert tuple(fn({"t": {"t": (1, 12)}})) == (12,)


def test_map_reduce_tree_async():
    assert (
        tree.map_reduce_tree(
            operator.second,
            lambda x, y: sum(x) + y,
            lambda x: operator.head(x) + 1,
        )((1, ((2, ()), (3, ()))))
        == 9
    )


async def test_map_reduce_tree():
    wait_time = 0.1

    async def increment(x):
        await asyncio.sleep(wait_time)
        return operator.head(x) + 1

    start_time = time.time()
    assert (
        await tree.map_reduce_tree(
            operator.second,
            lambda x, y: sum(x) + y,
            increment,
        )(
            (1, ((2, ()), (3, ()))),
        )
    ) == 9
    assert time.time() - start_time < wait_time * 1.1



import time

from gamla import immutable_set


def test_add():
    assert immutable_set.equals(
        immutable_set.add(
            immutable_set.create([1, 2, 3]),
            4,
        ),
        immutable_set.create(
            [1, 2, 3, 4],
        ),
    )


def test_remove():
    assert immutable_set.equals(
        immutable_set.remove(
            immutable_set.create([1, 2, 3]),
            2,
        ),
        immutable_set.create([1, 3]),
    )


def test_contains():
    assert immutable_set.contains(immutable_set.create([1, 2, 3]), 3)


def test_not_contains():
    assert not immutable_set.contains(immutable_set.create([1, 2, 3]), 4)


def test_union():
    assert immutable_set.equals(
        immutable_set.union(
            immutable_set.create([1, 2, 3, 4]),
            immutable_set.create([1, 2, 3]),
        ),
        immutable_set.create([1, 2, 3, 4]),
    )


def _is_o_of_1(f, arg1, arg2):
    start = time.perf_counter()
    f(arg1, arg2)
    return time.perf_counter() - start < 0.0001


_large_number = 9999


def test_intersection():
    assert immutable_set.equals(
        immutable_set.intersection(
            immutable_set.create([1, 2]),
            immutable_set.create([2]),
        ),
        immutable_set.create([2]),
    )


def test_performance_sanity():
    assert not _is_o_of_1(
        immutable_set.union,
        immutable_set.create(range(_large_number)),
        immutable_set.create(range(_large_number)),
    )


def test_union_performance():
    assert _is_o_of_1(
        immutable_set.union,
        immutable_set.create(range(_large_number)),
        immutable_set.create(range(_large_number // 64, _large_number // 32)),
    )


def test_intersection_performance():
    assert _is_o_of_1(
        immutable_set.intersection,
        immutable_set.create(range(_large_number)),
        immutable_set.create(range(1)),
    )



from gamla import construct, functional_generic


def test_wrap_str():
    assert functional_generic.pipe("john", construct.wrap_str("hi {}")) == "hi john"


def test_wrap_multiple_str():
    assert (
        functional_generic.pipe(
            {"first": "happy", "second": "world"},
            construct.wrap_multiple_str("hello {first} {second}"),
        )
        == "hello happy world"
    )


def test_wrap_dict():
    assert construct.wrap_dict("some_key")("some_value") == {"some_key": "some_value"}



from typing import Iterable

import immutables

# Design choices:
# - Using a class to allow for typing
# - The class has no methods to ensure all logic is in the functions below.
# - The wrapped map is kept private.
# - To prevent the user from making subtle mistake, we override `__eq__` to raise an error.
# Corollaries:
# - Will not work with operators ootb, e.g. `in`, `==` or `len`.


class ImmutableSet:
    def __init__(self, inner):
        self._inner = inner

    def __eq__(self, _):
        raise NotImplementedError(
            "Use the functions in this module instead of operators.",
        )


def create(iterable: Iterable) -> ImmutableSet:
    return ImmutableSet(immutables.Map(map(lambda x: (x, None), iterable)))


EMPTY: ImmutableSet = create([])


def equals(s1: ImmutableSet, s2: ImmutableSet) -> bool:
    return s1._inner == s2._inner  # noqa: SF01


def length(set: ImmutableSet) -> int:
    return len(set._inner)  # noqa: SF01


def add(set: ImmutableSet, element) -> ImmutableSet:
    return ImmutableSet(set._inner.set(element, None))  # noqa: SF01


def remove(set: ImmutableSet, element) -> ImmutableSet:
    return ImmutableSet(set._inner.delete(element))  # noqa: SF01


def contains(set: ImmutableSet, element) -> bool:
    return element in set._inner  # noqa: SF01


def union(set1: ImmutableSet, set2: ImmutableSet) -> ImmutableSet:
    smaller, larger = sorted([set1, set2], key=length)
    return ImmutableSet(larger._inner.update(smaller._inner))  # noqa: SF01


def intersection(set1: ImmutableSet, set2: ImmutableSet) -> ImmutableSet:
    smaller, larger = sorted([set1, set2], key=length)
    for element in smaller._inner:  # noqa: SF01
        if not contains(larger, element):
            smaller = remove(smaller, element)
    return smaller



from gamla import apply_utils, asyncio, operator


async def _opposite_async(x):
    await asyncio.sleep(0.01)
    return not x


def test_apply_method():
    class SomeClass:
        x: int

        def __init__(self, x):
            self.x = x

        def add(self, y):
            return self.x + y

    assert apply_utils.apply_method("add", 1)(SomeClass(2)) == 3


async def test_apply_method_async():
    class SomeClass:
        x: int

        def __init__(self, x):
            self.x = x

        async def mult_async(self, y):
            await asyncio.sleep(0.01)
            return self.x * y

    assert await apply_utils.apply_method_async("mult_async", 2)(SomeClass(2)) == 4


def test_apply():
    assert apply_utils.apply(1)(operator.add(2)) == 3


async def test_apply_async():
    assert not await apply_utils.apply_async(True)(_opposite_async)


async def test_apply_fn_with_args():
    assert apply_utils.apply_fn_with_args(lambda x, y: x + y, 1, 3) == 4



from typing import Dict, Text
from urllib import parse

from gamla import currying
from gamla.optimized import sync


@currying.curry
def add_to_query_string(params_to_add: Dict, url: Text) -> Text:
    """Add params_to_add to the query string part of url

    >>> add_to_query_string({ "param1" : "value"}, "http://domain.com")
    http://domain.com?param1=value
    """
    (scheme, netloc, path, query, fragment) = parse.urlsplit(url)
    return parse.urlunsplit(
        (
            scheme,
            netloc,
            path,
            parse.urlencode(
                sync.merge([parse.parse_qs(query), params_to_add]),
                doseq=True,
            ),
            fragment,
        ),
    )



import itertools
from typing import Any, Callable, Dict, FrozenSet, Iterable, Set, Text, Tuple

from gamla import construct, currying, dict_utils, functional_generic, operator
from gamla.optimized import sync


@currying.curry
def graph_traverse(
    source: Any,
    get_neighbors: Callable,
    key: Callable = operator.identity,
) -> Iterable:
    """Gets a graph and a function to get a node's neighbours, BFS over it from a single source node, return an iterator of unique nodes.

    >>> g = {'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}
    >>> list(graph_traverse('1', g.__getitem__))
    ['1', '2', '3', '4']"""
    yield from graph_traverse_many([source], get_neighbors=get_neighbors, key=key)


@currying.curry
def graph_traverse_many(
    sources: Any,
    get_neighbors: Callable,
    key: Callable = operator.identity,
) -> Iterable:
    """Gets a graph and a function to get a node's neighbours, BFS over it starting from multiple sources, return an iterator of unique nodes.

    >>> g = {'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}
    >>> list(graph_traverse_many(['1', '3'], g.__getitem__))
    ['3', '1', '4', '2']

    Note: `get_neighbors` must return elements without duplicates."""
    seen_set: Set = set()
    remember = functional_generic.compose_left(key, seen_set.add)
    should_traverse = functional_generic.compose_left(
        key,
        sync.complement(operator.contains(seen_set)),
    )
    yield from general_graph_traverse_many(
        sources,
        get_neighbors,
        remember,
        should_traverse,
    )


@currying.curry
def general_graph_traverse_many(
    sources: Any,
    get_neighbors: Callable,
    remember: Callable,
    should_traverse: Callable,
) -> Iterable:
    """Gets a graph, a function to get a node's neighbours, a function to add to the set of seen nodes, and function to know if a node is seen or not.
    BFS over the graph and return an iterator of unique nodes.

    >>> seen_set = set(); key =  len; g = {'one': ['two', 'three'], 'two': ['three'], 'three': ['four'], 'four': []}
    >>> list(general_graph_traverse_many(['one', 'three'], g.__getitem__, functional_generic.compose_left(key, seen_set.add), functional_generic.compose_left(key, sync.complement(operator.contains(seen_set)))))
    ['three', 'one', 'four']

    Note: `get_neighbors` must return elements without duplicates."""
    queue = [*sources]
    for element in queue:
        remember(element)

    while queue:
        current = queue.pop()
        yield current
        for node in get_neighbors(current):
            if should_traverse(node):
                remember(node)
                queue = [node] + queue


def traverse_graph_by_radius(
    source: Any,
    get_neighbors: Callable,
    radius: int,
) -> Iterable:
    """Traverse over a graph like `graph_traverse`, but up to a given radius.

    >>> g = {'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}
    >>> list(traverse_graph_by_radius('1', g.__getitem__, 1))
    ['1', '2', '3']"""

    def get_neighbors_limiting_radius(
        current_and_distance: Tuple[Text, int],
    ) -> Iterable[Tuple[Text, int]]:
        current, distance = current_and_distance
        if distance < radius:
            yield from map(
                lambda neighbor: (neighbor, distance + 1),
                get_neighbors(current),
            )

    return map(
        operator.head,
        graph_traverse(source=(source, 0), get_neighbors=get_neighbors_limiting_radius),
    )


#: Gets a sequence of edges and returns a graph made of these edges.
#:
#: >>> graph.edges_to_graph([(1,2), (2, 3), (3, 1), (3, 2)])
#: {1: frozenset({2}), 2: frozenset({3}), 3: frozenset({1, 2})}
edges_to_graph = functional_generic.compose(
    functional_generic.sync.valmap(
        functional_generic.sync.compose(
            frozenset,
            functional_generic.sync.map(operator.second),
        ),
    ),
    sync.groupby(operator.head),
)
#: Gets a graph and returns an iterator of all edges in it.
#:
#: >>> list(graph_to_edges({'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []}))
#: [('1', '2'), ('1', '3'), ('2', '3'), ('3', '4')]
graph_to_edges = functional_generic.compose_left(
    sync.keymap(construct.wrap_tuple),
    dict.items,
    sync.mapcat(sync.star(itertools.product)),
)

#: Gets a graph and returns the graph with its edges reversed
#:
#: >>> reverse_graph({'1': ['2', '3'], '2': ['3'], '3': ['4'], '4': []})
#: {'2': frozenset({'1'}), '3': frozenset({'1', '2'}), '4': frozenset({'3'})}
reverse_graph = functional_generic.compose_left(
    graph_to_edges,
    functional_generic.curried_map(functional_generic.compose_left(reversed, tuple)),
    edges_to_graph,
)

#: Gets a sequence of nodes (cliques) and returns the bidirectional graph they represent
#:
#: >>> cliques_to_graph([{1, 2}, {3, 4}])
#: {1: frozenset({2}), 2: frozenset({1}), 3: frozenset({4}), 4: frozenset({3})}
cliques_to_graph = functional_generic.compose_left(
    sync.mapcat(lambda clique: itertools.permutations(clique, r=2)),
    edges_to_graph,
)


def get_connectivity_components(graph: Dict) -> Iterable[FrozenSet]:
    """
    Gets a graph and return an iterator of its connectivity components.

    >>> g = cliques_to_graph([{1, 2}, {3, 4}])
    >>> list(get_connectivity_components(g))
    [frozenset({1, 2}), frozenset({3, 4})]

    Note: Graph is assumed to undirected, so each edge must appear both ways."""
    nodes_left = frozenset(graph)
    while nodes_left:
        result = frozenset(
            graph_traverse(
                source=operator.head(nodes_left),
                get_neighbors=functional_generic.compose(
                    sync.filter(operator.contains(nodes_left)),
                    graph.get,
                ),
            ),
        )
        yield result
        nodes_left -= result


@currying.curry
def groupby_many(f: Callable, it: Iterable) -> Dict[Text, Any]:
    """Return a mapping `{y: {x s.t. y in f(x)}}, where x in it. `

    Parameters:
    Key function (gets an object in collection and outputs tuple of keys).
    A Collection.

    Returns a dictionary where key has been computed by the `f` key function.

    >>> names = ['alice', 'bob', 'charlie', 'dan', 'edith', 'frank']
    >>> groupby_many(lambda name: (name[0], name[-1]), names)
    {'a': frozenset({'alice'}),
     'e': frozenset({'alice', 'charlie', 'edith'}),
     'b': frozenset({'bob'}),
     'c': frozenset({'charlie'}),
     'd': frozenset({'dan'}),
     'n': frozenset({'dan'}),
     'h': frozenset({'edith'}),
     'f': frozenset({'frank'}),
     'k': frozenset({'frank'})}"""
    return functional_generic.pipe(
        it,
        functional_generic.mapcat(
            functional_generic.compose_left(
                lambda element: (f(element), [element]),
                sync.star(itertools.product),
            ),
        ),
        edges_to_graph,
    )


@currying.curry
def _has_cycle(sourced, get_neighbors, visited, node):
    if node in sourced:
        return True
    if node in visited:
        return False
    visited.add(node)
    return functional_generic.pipe(
        node,
        get_neighbors,
        functional_generic.anymap(_has_cycle(sourced | {node}, get_neighbors, visited)),
    )


def has_cycle(graph):
    """Gets a graph, returns True if it contains a cycle, False else.

    >>> has_cycle({1: [2], 2: [3], 3: [1]})
    True
    >>> has_cycle({1: [2], 2: [3]})
    False"""
    return functional_generic.pipe(
        graph,
        dict.keys,
        functional_generic.curried_map(
            _has_cycle(
                frozenset(),
                dict_utils.dict_to_getter_with_default((), graph),
                set(),
            ),
        ),
        any,
    )


_non_sources = sync.compose_left(
    dict.values,
    operator.concat,
    frozenset,
)


def find_sources(graph: Dict) -> FrozenSet:
    """Gets a directional graph and returns its sources.

    >>> find_sources({'1': ['2', '3'], '2': ['3'], '3': [], '4': []})
    frozenset({'1', '4'})
    """
    return sync.pipe(
        graph,
        sync.remove(operator.contains(_non_sources(graph))),
        frozenset,
    )



from gamla import functional_generic, operator
from gamla.optimized import async_functions


def prepare_and_apply(f):
    """Transforms a higher order function to a regular one.

    Uses the given value once to prepare a regular function, then again to call it with.

    >>> def increment(x): return x + 1
    >>> def decrement(x): return x - 1
    >>> def conditional_transformation(x):
    ...     return increment if x < 10 else decrement
    >>> prepare_and_apply(conditional_transformation))(15)
    14
    """

    def prepare_and_apply(value):
        return f(value)(value)

    return prepare_and_apply


def prepare_and_apply_async(f):
    """Transforms a higher order function to a regular one.

    Uses the given value once to prepare a regular function, then again to call it with.

    >>> async def increment(x): return x + 1
    >>> async def decrement(x): return x - 1
    >>> def conditional_transformation(x):
    ...     return increment if x < 10 else decrement
    >>> prepare_and_apply(conditional_transformation))(15)
    14
    """

    async def prepare_and_apply(value):
        return await async_functions.to_awaitable(
            (await async_functions.to_awaitable(f(value)))(value),
        )

    return prepare_and_apply


#: Make a function act on the first element on incoming input.
on_first = functional_generic.before(operator.head)
#: Make a function act on the second element on incoming input.
on_second = functional_generic.before(operator.second)



from gamla import functional_generic, transducer


def _append_to_tuple(s, current):
    """Highly inefficient, but just for testing purposes."""
    return (*s, current)


_increment = transducer.map(lambda x: x + 1)


def test_map():
    assert transducer.transduce(_increment, _append_to_tuple, [], [1, 2, 3]) == (
        2,
        3,
        4,
    )


def test_filter():
    assert transducer.transduce(
        transducer.filter(lambda x: x > 2),
        _append_to_tuple,
        [],
        [1, 2, 3],
    ) == (3,)


def test_composition():
    assert transducer.transduce(
        functional_generic.compose(_increment, _increment),
        _append_to_tuple,
        (),
        [1, 2, 3],
    ) == (
        3,
        4,
        5,
    )


def test_composition_of_3_functions():
    assert (
        transducer.transduce(
            functional_generic.compose(
                _increment,
                _increment,
                lambda _: lambda s, x: x + s,
            ),
            lambda s, _: s,
            0,
            [1, 2, 3],
        )
        == 12
    )


def test_juxt():
    assert transducer.transduce(
        transducer.juxt(
            _increment(_append_to_tuple),
            lambda s, x: x + s,
        ),
        lambda s, _: s,
        ((), 0),
        [1, 2, 3],
    ) == ((2, 3, 4), 6)


def test_apply_spec():
    assert transducer.transduce(
        transducer.apply_spec(
            {
                "incremented": _increment(_append_to_tuple),
                "sum": lambda s, x: x + s,
            },
        ),
        lambda s, _: s,
        {"incremented": (), "sum": 0},
        [1, 2, 3],
    ) == {"incremented": (2, 3, 4), "sum": 6}


def test_count_by():
    assert transducer.transduce(
        transducer.count_by(
            lambda x: ("even",) if x % 2 == 0 else ("odd",),
        ),
        lambda s, _: s,
        {},
        [1, 2, 3, 4, 5],
    ) == {"even": 2, "odd": 3}
